import dash

from webviz_ert.models import EnsembleModel, load_ensemble
from webviz_ert.plugins import ObservationAnalyzer, ParameterComparison


def test_ensemble_model(mock_data):
    ens_model = EnsembleModel(ensemble_id=1, project_id=None)
    assert ens_model.name == "default"
    # Will be same as responses in experiment
    assert len(ens_model.responses) == 4


def test_ensemble_model_parameter_data(mock_data):
    ens_id = 42
    ens_model = EnsembleModel(ensemble_id=ens_id, project_id=None)
    parameters = ens_model.parameters
    assert len(parameters) == 2
    assert parameters["OP1_DIVERGENCE_SCALE"].data_df().values.flatten().tolist() == [
        0.1,
        1.1,
        2.1,
    ]
    assert parameters["BPR_138_PERSISTENCE"].data_df().values.flatten().tolist() == [
        0.01,
        1.01,
        2.01,
    ]


def test_ensemble_caching(mock_data):
    app = dash.Dash(__name__)

    plotter_view = ParameterComparison(app, project_identifier=None)
    ens_model_1 = EnsembleModel(ensemble_id=1, project_id=None)
    assert len(plotter_view.get_ensembles()) == 0

    ensemble = plotter_view.get_ensemble(ensemble_id=1)
    assert ensemble is None

    plotter_view.add_ensemble(ens_model_1)
    load_ensemble(plotter_view, ensemble_id=42)

    plotter_view_ensembles = plotter_view.get_ensembles()

    assert len(plotter_view_ensembles) == 2
    assert [ens_id for ens_id, ens_model in plotter_view_ensembles.items()] == [
        1,
        42,
    ]

    page_obs_analyzer = ObservationAnalyzer(app, project_identifier=None)
    page_obs_analyzer_ensembles = page_obs_analyzer.get_ensembles()
    assert len(page_obs_analyzer_ensembles) == 2
    assert page_obs_analyzer_ensembles == plotter_view_ensembles

    plotter_view.clear_ensembles()
    assert len(page_obs_analyzer_ensembles) == 0
    assert len(plotter_view_ensembles) == 0



import dash
import plotly.graph_objects as go
from typing import Any, List, Optional, Dict, Tuple

from dash.dependencies import Input, Output, State
from webviz_ert.models import (
    load_ensemble,
    ParallelCoordinatesPlotModel,
)
from webviz_ert.plugins import WebvizErtPluginABC
from webviz_ert import assets


def parameter_comparison_controller(parent: WebvizErtPluginABC, app: dash.Dash) -> None:
    graph_id = {"id": parent.uuid("parallel-coor"), "type": parent.uuid("graph")}

    @app.callback(
        [Output(graph_id, "figure"), Output(graph_id, "style")],
        [
            Input(parent.uuid("parameter-selection-store-param"), "modified_timestamp"),
            Input(parent.uuid("ensemble-selection-store"), "modified_timestamp"),
        ],
        [
            State(parent.uuid("parameter-selection-store-param"), "data"),
            State(parent.uuid("ensemble-selection-store"), "data"),
        ],
    )
    def update_parallel_coor(
        _: Any,
        __: Any,
        selected_parameters: Optional[List[str]],
        selected_ensembles: Optional[Dict[str, List]],
    ) -> Tuple[go.Figure, Optional[Dict[str, str]]]:
        ensembles = None
        if selected_ensembles and selected_ensembles["selected"]:
            ensembles = selected_ensembles["selected"]

        # If no ensemble or parameter is selected just don't display the figure
        if not ensembles or not selected_parameters:
            return go.Figure(), {"display": "none"}

        data = {}
        colors = {}
        for idx, selected_ensemble in enumerate(ensembles):
            ensemble = load_ensemble(parent, selected_ensemble["value"])
            ens_key = ensemble.name
            df = ensemble.parameters_df(selected_parameters)
            df["ensemble_id"] = idx
            data[ens_key] = df.copy()
            colors[ens_key] = assets.get_color(index=idx)
        parallel_plot = ParallelCoordinatesPlotModel(data, colors)

        return parallel_plot.repr, None



import pandas as pd
import numpy as np
from webviz_ert.controllers.multi_response_controller import (
    _get_observation_plots,
    _get_realizations_plots,
    _get_realizations_statistics_plots,
)
from webviz_ert.controllers.observation_response_controller import (
    _get_univariate_misfits_boxplots,
)

from webviz_ert.data_loader import get_ensembles
from webviz_ert.models import EnsembleModel, PriorModel
from webviz_ert.models import (
    HistogramPlotModel,
    MultiHistogramPlotModel,
    BoxPlotModel,
    ParallelCoordinatesPlotModel,
    BarChartPlotModel,
)
import webviz_ert.assets as assets


def test_observation_plot_representation():
    observation_df = pd.DataFrame(
        data={
            "values": [2.85325093, 7.20311703, 21.38648991, 31.51455593, 53.56766604],
            "std": [0.1, 1.1, 4.1, 9.1, 16.1],
            "x_axis": [0, 2, 4, 6, 8],
            "attributes": "Key1 Value1<br>Key2 Value2<br>",
            "active": [True, False, True, False, False],
        }
    )

    plot = _get_observation_plots(observation_df)

    assert "mode" in plot.repr

    np.testing.assert_equal(observation_df["x_axis"].values, plot.repr.x)
    assert len(plot.repr.y) == len(observation_df)

    np.testing.assert_equal(plot.repr.y, observation_df["values"].values)
    np.testing.assert_equal(plot.repr.error_y.array, observation_df["std"].values)
    np.testing.assert_equal(plot.repr.text, observation_df["attributes"].values)
    np.testing.assert_equal(
        plot.repr.marker.color,
        (
            "rgb(176, 28, 52)",
            "rgb(0, 0, 0)",
            "rgb(176, 28, 52)",
            "rgb(0, 0, 0)",
            "rgb(0, 0, 0)",
        ),
    )


def test_realizations_plot_representation():
    data = np.random.rand(200).reshape(-1, 20)
    realization_df = pd.DataFrame(data=data, index=range(10), columns=range(20))
    x_axis = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    plots = _get_realizations_plots(
        realization_df, x_axis, assets.ERTSTYLE["ensemble-selector"]["color_wheel"][0]
    )
    assert len(plots) == 20
    for idx, plot in enumerate(plots):
        np.testing.assert_equal(x_axis, plot.repr.x)
        np.testing.assert_equal(plot.repr.y, realization_df[idx].values)


def test_realizations_statistics_plot_representation():
    data = np.random.rand(200).reshape(-1, 20)
    realization_df = pd.DataFrame(data=data, index=range(10), columns=range(20))
    x_axis = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    plots = _get_realizations_statistics_plots(
        realization_df, x_axis, assets.ERTSTYLE["ensemble-selector"]["color_wheel"][0]
    )

    assert len(plots) == 3
    assert "mode" in plots[0].repr
    for plot in plots:
        np.testing.assert_equal(x_axis, plot.repr.x)

    np.testing.assert_equal(plots[0].repr.y, np.mean(data, axis=1))
    np.testing.assert_equal(plots[1].repr.y, np.quantile(data, 0.1, axis=1))
    np.testing.assert_equal(plots[2].repr.y, np.quantile(data, 0.9, axis=1))


def test_histogram_plot_representation():
    data = np.random.rand(20).reshape(-1, 20)
    data_df = pd.DataFrame(data=data, index=range(1), columns=range(20))
    data_df.index.name = "key_name"

    plot = HistogramPlotModel(data_df, hist=True, kde=False)
    plot.selection = range(5)
    plot = plot.repr
    np.testing.assert_equal(plot.data[0].x, data.flatten()[:5])
    assert plot.data[0].histnorm == "probability density"
    assert plot.data[0].autobinx == False


def test_multi_histogram_plot_representation():
    data_dict = {}
    name_dict = {}
    colors_dict = {}

    ensemble_names = ["default", "update_1", "update_2"]
    colors = assets.ERTSTYLE["ensemble-selector"]["color_wheel"]
    for ensemble_date, (ensemble_name, color) in enumerate(
        zip(ensemble_names, colors[: len(ensemble_names)])
    ):
        key = f"{ensemble_date}, {ensemble_name}"
        data = np.random.rand(20).reshape(-1, 20)
        data_df = pd.DataFrame(data=data, index=range(1), columns=range(20))
        data_df.index.name = "KEY_NAME"
        data_dict[key] = data_df
        name_dict[key] = f"{ensemble_name}"
        colors_dict[key] = color
    priors = {
        (0, "default"): (PriorModel("uniform", ["std", "mean"], [0, 1]), colors[0])
    }

    plot = MultiHistogramPlotModel(
        data_dict, name_dict, colors_dict, hist=True, kde=False
    )
    assert plot.bin_count == 4

    plot = MultiHistogramPlotModel(
        data_dict,
        name_dict,
        colors_dict,
        hist=True,
        kde=False,
        priors=priors,
        bin_count=10,
    )
    assert plot.bin_count == 10
    plot = plot.repr
    for idx, ensemble_name in enumerate(ensemble_names):
        key = f"{idx}, {ensemble_name}"
        np.testing.assert_equal(plot.data[idx].x, data_dict[key].values.flatten())
        assert plot.data[idx].histnorm == "probability density"
        assert plot.data[idx].autobinx == False
        assert plot.data[idx].marker.color == colors_dict[key]
        assert plot.data[idx].name == ensemble_name

    assert plot.data[-1].name == "(0, 'default')-prior"


def test_parallel_coordinates_representation():
    data_dict = {}
    colors_dict = {}

    ensemble_names = ["default", "update_1", "update_2"]
    colors = assets.ERTSTYLE["ensemble-selector"]["color_wheel"]
    for idx, (ensemble_name, color) in enumerate(
        zip(ensemble_names, colors[: len(ensemble_names)])
    ):
        key = f"{idx}, {ensemble_name}"
        data = np.random.rand(50).reshape(-1, 5)
        data_df = pd.DataFrame(data=data, columns=[f"PARAM_{i}" for i in range(5)])
        data_df["ensemble_id"] = idx
        data_dict[key] = data_df
        colors_dict[key] = color

    plot = ParallelCoordinatesPlotModel(data_dict, colors_dict)
    plot = plot.repr

    assert len(plot.data[0].dimensions) == 5
    for idx, ensemble_name in enumerate(ensemble_names):
        key = f"{idx}, {ensemble_name}"
        assert plot.data[0].dimensions[idx].label == f"PARAM_{idx}"
        assert len(plot.data[0].dimensions[idx].values) == 10 * len(ensemble_names)
        assert plot.data[0].labelangle == 45
        assert len(plot.data[0].line.color) == 10 * len(ensemble_names)


def test_univariate_misfits_boxplot_representation():
    data = np.random.rand(200).reshape(-1, 20)
    missfits_df = pd.DataFrame(data=data, index=range(10), columns=range(20))
    ensemble_name = "test-ensemble"
    plots = _get_univariate_misfits_boxplots(
        missfits_df.copy(), ensemble_name=ensemble_name, color="rgb(255,0,0)"
    )
    assert len(plots) == 20
    for id_plot, plot in enumerate(plots):
        np.testing.assert_equal(0.3, plot.repr.jitter)
        np.testing.assert_equal("all", plot.repr.boxpoints)
        x_pos = missfits_df.columns[id_plot]
        if isinstance(x_pos, int):
            name = f"Value {x_pos}"
        else:
            name = f"{x_pos} - {ensemble_name}"
        assert name == plot.repr.name


def test_boxplot_representation():
    data = np.random.rand(10)
    data_df = pd.DataFrame(data=data, index=range(10))

    plot = BoxPlotModel(
        y_axis=data_df.values,
        name="Boxplot@Location5",
        color=assets.ERTSTYLE["ensemble-selector"]["color_wheel"][0],
    )
    plot = plot.repr
    np.testing.assert_equal(plot.y.flatten(), data)
    assert plot.boxpoints == "all"
    assert plot.name == "Boxplot@Location5"


def test_barchart_representation():
    data_dict = {}
    colors_dict = {}

    ensemble_names = ["default", "update_1", "update_2"]
    colors = assets.ERTSTYLE["ensemble-selector"]["color_wheel"]
    param_num = 5
    for idx, (ensemble_name, color) in enumerate(
        zip(ensemble_names, colors[: len(ensemble_names)])
    ):
        key = f"{idx}, {ensemble_name}"
        data = np.random.rand(param_num)
        data_df = pd.DataFrame(
            data=data, index=[f"PARAM_{i}" for i in range(param_num)]
        )
        data_dict[key] = data_df
        colors_dict[key] = color

    plot = BarChartPlotModel(data_dict, colors_dict)
    plot = plot.repr

    for idx, ensemble_name in enumerate(ensemble_names):
        key = f"{idx}, {ensemble_name}"
        assert plot.data[idx].name == key
        np.testing.assert_equal(
            plot.data[idx].y, [f"PARAM_{i}" for i in range(param_num)]
        )
        assert plot.data[idx].orientation == "h"
        assert plot.data[idx].x.shape == (param_num, 1)
        assert plot.data[idx].marker.color == colors[idx]



from typing import List
from dash.development.base_component import Component
from webviz_ert.plugins import WebvizErtPluginABC

from dash import html
from dash import dcc
import webviz_core_components as wcc
import dash_bootstrap_components as dbc
from .selector_view import parameter_selector_view
from webviz_ert.models.data_model import DataType
from .ensemble_selector_view import ensemble_selector_list


def plot_view_header(parent: WebvizErtPluginABC) -> List[Component]:
    return [
        dbc.Row(
            [
                dbc.Col(
                    id=parent.uuid("ensemble-content"),
                    children=ensemble_selector_list(parent=parent),
                    width=4,
                ),
                dbc.Col(
                    [
                        parameter_selector_view(
                            parent,
                            data_type=DataType.RESPONSE,
                            titleLabel="Responses",
                        ),
                        dcc.Checklist(
                            id=parent.uuid("response-observations-check"),
                            options=[
                                {
                                    "label": "Show only responses with observations",
                                    "value": "obs",
                                },
                                {
                                    "label": "Remove all key-names ending with 'H' ( "
                                    "probably historical vectors )",
                                    "value": "historical",
                                },
                            ],
                            value=[],
                            labelStyle={"display": "block"},
                        ),
                    ],
                    width=4,
                    id=parent.uuid("response-section"),
                ),
                dbc.Col(
                    [
                        parameter_selector_view(
                            parent,
                            data_type=DataType.PARAMETER,
                            titleLabel="Parameters",
                        ),
                    ],
                    width=4,
                    id=parent.uuid("parameter-section"),
                ),
            ]
        ),
        dcc.Store(id=parent.uuid("plot-selection-store-resp"), storage_type="session"),
        dcc.Store(id=parent.uuid("plot-selection-store-param"), storage_type="session"),
    ]


def plot_view_body(parent: WebvizErtPluginABC) -> List[Component]:
    return [
        html.Div(
            [
                dbc.Row(children=[], id=parent.uuid("plotting-content-resp")),
                dbc.Row(children=[], id=parent.uuid("plotting-content-param")),
            ],
            id=parent.uuid("plotting-content-container"),
        ),
    ]


def plot_view_menu(parent: WebvizErtPluginABC) -> List[Component]:
    return [
        html.Div(
            dcc.Checklist(
                id=parent.uuid("param-label-check"),
                options=[
                    {
                        "label": "Show legend description",
                        "value": "label",
                    }
                ],
                value=["label"],
                persistence="session",
            ),
            className="ert-parameter-label-checkbox",
        ),
    ]



from typing import List, Mapping, Optional, Any
import pandas as pd
from webviz_ert.data_loader import get_data_loader, DataLoader

from webviz_ert.models import Observation, AxisType


class Response:
    def __init__(
        self,
        name: str,
        ensemble_id: str,
        project_id: str,
        ensemble_size: int,
        active_realizations: List[int],
        resp_schema: Any,
    ):
        self._data_loader: DataLoader = get_data_loader(project_id)
        self._document: Optional[Mapping[str, Any]] = None
        self._id: str = resp_schema["id"]
        self._ensemble_id: str = ensemble_id
        self.name: str = name
        self._data: Optional[pd.DataFrame] = None
        self._observations: Optional[List[Observation]] = None
        self._univariate_misfits_df: Optional[pd.DataFrame] = None
        self._summary_misfits_df: Optional[pd.DataFrame] = None
        self._ensemble_size: int = ensemble_size
        self._active_realizations: List[int] = active_realizations
        self._has_observations: bool = resp_schema.get("has_observations")

    @property
    def ensemble_id(self) -> str:
        return self._ensemble_id

    @property
    def axis(self) -> pd.Index:
        return self.data.index

    @property
    def axis_type(self) -> Optional[AxisType]:
        if self.axis is None or self.axis.empty:
            return None
        if str(self.axis[0]).isnumeric():
            return AxisType.INDEX
        return AxisType.TIMESTAMP

    @property
    def data(self) -> pd.DataFrame:
        if self._data is None:
            self._data = self._data_loader.get_ensemble_record_data(
                self._ensemble_id, self.name
            )
        return self._data

    @property
    def empty(self) -> bool:
        return self.data.empty

    def univariate_misfits_df(
        self, selection: Optional[List[int]] = None
    ) -> pd.DataFrame:
        if self._univariate_misfits_df is None:
            self._univariate_misfits_df = self._data_loader.compute_misfit(
                self._ensemble_id, self.name, summary=False
            )
        if selection:
            return self._univariate_misfits_df.iloc[selection, :]
        return self._univariate_misfits_df

    def summary_misfits_df(self, selection: Optional[List[int]] = None) -> pd.DataFrame:
        if self._summary_misfits_df is None:
            self._summary_misfits_df = self._data_loader.compute_misfit(
                self._ensemble_id, self.name, summary=True
            )
        if selection:
            self._summary_misfits_df.iloc[selection, :]
        return self._summary_misfits_df

    def data_df(self, selection: Optional[List[int]] = None) -> pd.DataFrame:
        if selection:
            self.data.iloc[selection, :]
        return self.data

    @property
    def observations(self) -> Optional[List[Observation]]:
        if self._observations is None:
            _observations_schemas = self._data_loader.get_ensemble_record_observations(
                self._ensemble_id, self.name
            )
            self._observations = []
            for observation_schema in _observations_schemas:
                self._observations.append(
                    Observation(observation_schema=observation_schema)
                )
        return self._observations

    @property
    def has_observations(self) -> bool:
        return self._has_observations



from ._webviz_ert import WebvizErtPluginABC
from ._response_comparison import ResponseComparison
from ._observation_analyzer import ObservationAnalyzer
from ._parameter_comparison import ParameterComparison
from ._response_correlation import ResponseCorrelation



import io

import pandas as pd

from webviz_ert.data_loader import escape


class DataContent:
    def __init__(self, content):
        self._content = content.encode()

    @property
    def content(self):
        return self._content


def to_parquet_helper(dataframe: pd.DataFrame) -> bytes:
    stream = io.BytesIO()
    dataframe.to_parquet(stream)
    return stream.getvalue()


all_ensemble_names = [
    "default",
    "default3",
    "default_smoother_update",
    "nr_42",
]

_experiment_1_metadata = {
    "name": "default",
    "id": 1,
    "ensemble_ids": [1, 2, 3, 42],
    "priors": {
        "SNAKE_OIL_PARAM:BPR_138_PERSISTENCE": {
            "function": "UNIFORM",
            "parameter_names": ["MIN", "MAX"],
            "parameter_values": [0.2, 0.7],
        },
        "SNAKE_OIL_PARAM:OP1_DIVERGENCE_SCALE": {
            "function": "UNIFORM",
            "parameter_names": ["MIN", "MAX"],
            "parameter_values": [0.2, 0.7],
        },
    },
    "parameters": {
        "OP1_DIVERGENCE_SCALE": {
            "type": "gen_kw",
            "name": "OP1_DIVERGENCE_SCALE",
            "dimensionality": 1,
            "forward_init": False,
            "update": True,
            "distribution": {"name": "uniform", "min": 0.0, "max": 1.0},
            "group": "SNAKE_OIL_PARAM",
            "input_source": "sampled",
        },
        "BPR_138_PERSISTENCE": {
            "type": "gen_kw",
            "name": "BPR_138_PERSISTENCE",
            "dimensionality": 1,
            "forward_init": False,
            "update": True,
            "distribution": {"name": "uniform", "min": 0.0, "max": 1.0},
            "group": "SNAKE_OIL_PARAM",
            "input_source": "sampled",
        },
    },
    "responses": {
        "summary": [
            {"response_type": "summary", "response_key": "FGPT", "filter_on": None},
            {"response_type": "summary", "response_key": "WOPR:OP1", "filter_on": None},
            {"response_type": "summary", "response_key": "FOPR", "filter_on": None},
        ],
        "gen_data": [
            {
                "response_type": "gen_data",
                "response_key": "SNAKE_OIL_GPR_DIFF",
                "filter_on": {"report_step": [199]},
            }
        ],
    },
    "observations": {
        "summary": {
            "FOPR": ["FOPR"],
            "WOPR:OP1": [
                "WOPR_OP1_108",
                "WOPR_OP1_9",
                "WOPR_OP1_144",
                "WOPR_OP1_190",
                "WOPR_OP1_36",
                "WOPR_OP1_72",
            ],
        }
    },
    "userdata": {},
}

ensembles_response = {
    "http://127.0.0.1:5000/updates/facade": "OK",
    "http://127.0.0.1:5000/experiments": [
        _experiment_1_metadata,
    ],
    "http://127.0.0.1:5000/experiments/1": _experiment_1_metadata,
    "http://127.0.0.1:5000/ensembles/1": {
        "child_ensemble_ids": [2],
        "experiment_id": 1,
        "parent_ensemble_id": None,
        "id": 1,
        "size": 1,
        "active_realizations": [0],
        "userdata": {"name": "default"},
    },
    "http://127.0.0.1:5000/ensembles/1/parameters": [
        {"name": "BPR_138_PERSISTENCE", "labels": []},
        {"name": "OP1_DIVERGENCE_SCALE", "labels": []},
    ],
    "http://127.0.0.1:5000/ensembles/1/responses": {
        "SNAKE_OIL_GPR_DIFF": {
            "name": "SNAKE_OIL_GPR_DIFF",
            "id": "SNAKE_OIL_GPR_DIFF",
            "has_observations": False,
        },
    },
    "http://127.0.0.1:5000/ensembles/2": {
        "experiment_id": 1,
        "child_ensemble_ids": [],
        "parent_ensemble_id": 1,
        "id": 2,
        "size": 1,
        "active_realizations": [0],
        "userdata": {"name": "default_smoother_update"},
    },
    "http://127.0.0.1:5000/ensembles/2/parameters": [
        {"name": "test_parameter_1", "labels": []},
        {"name": "test_parameter_11", "labels": []},
        {"name": "BPR_138_PERSISTENCE", "labels": []},
    ],
    "http://127.0.0.1:5000/ensembles/2/responses": {
        "SNAKE_OIL_GPR_DIFF": {
            "name": "SNAKE_OIL_GPR_DIFF",
            "id": "SNAKE_OIL_GPR_DIFF",
        },
    },
    "http://127.0.0.1:5000/ensembles/3": {
        "experiment_id": 1,
        "child_ensemble_ids": [],
        "name": "default3",
        "parent_ensemble_id": None,
        "id": 3,
        "size": 1,
        "active_realizations": [0],
        "userdata": {"name": "default3"},
    },
    "http://127.0.0.1:5000/ensembles/3/responses": {
        "SNAKE_OIL_GPR_DIFF": {
            "name": "SNAKE_OIL_GPR_DIFF",
            "id": "SNAKE_OIL_GPR_DIFF",
            "has_observations": False,
        },
        "FGPT": {
            "name": "FGPT",
            "id": "FGPT",
            "has_observations": False,
        },
        "FOPR": {
            "name": "FOPR",
            "id": "FOPR",
            "has_observations": True,
        },
        "WOPR:OP1": {
            "name": "WOPR:OP1",
            "id": "WOPR:OP1",
            "has_observations": True,
        },
    },
    "http://127.0.0.1:5000/ensembles/3/parameters": [
        {"name": "BPR_138_PERSISTENCE", "labels": []},
        {"name": "OP1_DIVERGENCE_SCALE", "labels": []},
    ],
    "http://127.0.0.1:5000/ensembles/4": {
        "experiment_id": 1,
        "child_ensemble_ids": [],
        "parent_ensemble_id": None,
        "id": 4,
        "size": 1,
        "active_realizations": [0],
        "userdata": {"name": "default4"},
    },
    "http://127.0.0.1:5000/ensembles/4/responses": {
        "SNAKE_OIL_GPR_DIFF": {
            "name": "SNAKE_OIL_GPR_DIFF",
            "id": "SNAKE_OIL_GPR_DIFF",
        },
        "FOPR": {
            "name": "FOPR",
            "id": "FOPR",
            "has_observations": True,
        },
    },
    "http://127.0.0.1:5000/ensembles/1/responses/SNAKE_OIL_GPR_DIFF/observations?realization_index=0": [],
    "http://127.0.0.1:5000/ensembles/3/responses/SNAKE_OIL_GPR_DIFF?realization_index=0": pd.DataFrame(
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        columns=[0],
        index=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    )
    .transpose()
    .to_csv()
    .encode(),
    "http://127.0.0.1:5000/ensembles/3/responses/SNAKE_OIL_GPR_DIFF/observations?realization_index=0": [],
    "http://127.0.0.1:5000/ensembles/3/responses/FOPR/observations?realization_index=0": [
        {
            "x_axis": ["2010-01-10 00:00:00", "2010-04-10 00:00:00"],
            "errors": [4, 2],
            "values": [0.42, 0.24],
            "name": "FOPR",
        }
    ],
    "http://127.0.0.1:5000/ensembles/3/responses/FOPR": to_parquet_helper(
        pd.DataFrame(
            [0.24, 0.13, 0.22, 0.36, 0.21, 0.54, 0.12, 0.16, 0.23, 0.18],
            index=[
                "2010-01-10 00:00:00",
                "2010-02-10 00:00:00",
                "2010-03-10 00:00:00",
                "2010-04-10 00:00:00",
                "2010-05-10 00:00:00",
                "2010-06-10 00:00:00",
                "2010-07-10 00:00:00",
                "2010-08-10 00:00:00",
                "2010-09-10 00:00:00",
                "2010-10-10 00:00:00",
            ],
            columns=[0],
        ).transpose()
    ),
    "http://127.0.0.1:5000/ensembles/3/responses/FGPT/observations?realization_index=0": [],
    "http://127.0.0.1:5000/ensembles/3/responses/FGPT": to_parquet_helper(
        pd.DataFrame(
            [0.1, 1.1, 2.1, 3.1, 4.1, 5.1, 6.1, 7.1, 8.1, 9.1],
            columns=[0],
            index=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
        ).transpose()
    ),
    "http://127.0.0.1:5000/ensembles/3/responses/WOPR%253AOP1": to_parquet_helper(
        pd.DataFrame(
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            columns=[0],
            index=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
        ).transpose()
    ),
    "http://127.0.0.1:5000/ensembles/3/responses/WOPR%253AOP1/observations?realization_index=0": [
        {
            "x_axis": [1, 4],
            "errors": [1, 1],
            "values": [1, 5],
            "name": "WOPR:OP1",
        }
    ],
    "http://127.0.0.1:5000/ensembles/4/responses/SNAKE_OIL_GPR_DIFF?realization_index=0": pd.DataFrame(
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        columns=[0],
        index=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    )
    .transpose()
    .to_csv()
    .encode(),
    "http://127.0.0.1:5000/ensembles/4/responses/SNAKE_OIL_GPR_DIFF/observations?realization_index=0": [],
    "http://127.0.0.1:5000/ensembles/4/responses/FOPR?realization_index=0": pd.DataFrame(
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        columns=[0],
        index=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    )
    .transpose()
    .to_csv()
    .encode(),
    "http://127.0.0.1:5000/ensembles/4/responses/FOPR/observations?realization_index=0": [
        {
            "x_axis": {"data": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]},
            "errors": {"data": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]},
            "values": {"data": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]},
            "name": "FOPR",
        }
    ],
}


ensembles_response[
    "http://127.0.0.1:5000/ensembles/1/responses/SNAKE_OIL_GPR_DIFF?realization_index=0"
] = to_parquet_helper(
    pd.DataFrame(
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        columns=["0"],
        index=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
    ).transpose()
)

ensembles_response[
    'http://127.0.0.1:5000/ensembles/1/responses/SNAKE_OIL_GPR_DIFF?filter_on={"report_step": "199"}'
] = to_parquet_helper(
    pd.DataFrame(
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        columns=["0"],
        index=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
    ).transpose()
)

ensembles_response[
    "http://127.0.0.1:5000/ensembles/1/parameters/OP1_DIVERGENCE_SCALE"
] = to_parquet_helper(
    pd.DataFrame(
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        columns=["0"],
        index=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
    ).transpose()
)


ensembles_response[
    "http://127.0.0.1:5000/ensembles/1/parameters/BPR_138_PERSISTENCE"
] = to_parquet_helper(
    pd.DataFrame(
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        columns=["0"],
        index=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
    ).transpose()
)

ensembles_response.update(
    {
        "http://127.0.0.1:5000/ensembles/42": {
            "child_ensemble_ids": [],
            "experiment_id": 1,
            "parent_ensemble_id": None,
            "id": 42,
            "name": "nr_42",
            "size": 1,
            "active_realizations": [0],
            "userdata": {"name": "nr_42"},
        },
        "http://127.0.0.1:5000/ensembles/42/parameters": [
            [
                {"name": "SNAKE_OIL_PARAM:OP1_DIVERGENCE_SCALE", "labels": []},
                {"name": "SNAKE_OIL_PARAM:BPR_138_PERSISTENCE", "labels": []},
            ]
        ],
        "http://127.0.0.1:5000/ensembles/42/responses": {
            "SNAKE_OIL_PARAM:OP1_DIVERGENCE_SCALE": {
                "name": "SNAKE_OIL_PARAM:BPR_138_PERSISTENCE",
                "id": "SNAKE_OIL_PARAM:OP1_DIVERGENCE_SCALE",
            },
            "SNAKE_OIL_PARAM:BPR_138_PERSISTENCE": {
                "name": "SNAKE_OIL_PARAM:BPR_138_PERSISTENCE",
                "id": "SNAKE_OIL_PARAM:BPR_138_PERSISTENCE",
            },
        },
    }
)


ensembles_response[
    f"http://127.0.0.1:5000/ensembles/42/parameters/OP1_DIVERGENCE_SCALE?"
] = to_parquet_helper(
    pd.DataFrame(
        [0.1, 1.1, 2.1],
        columns=["0"],
        index=["0", "1", "2"],
    ).transpose()
)
ensembles_response[
    f"http://127.0.0.1:5000/ensembles/42/parameters/BPR_138_PERSISTENCE?"
] = to_parquet_helper(
    pd.DataFrame(
        [0.01, 1.01, 2.01],
        columns=["a"],
        index=["0", "1", "2"],
    ).transpose()
)



from typing import List, Union, Optional, TYPE_CHECKING
import datetime
import pandas as pd


def indexes_to_axis(
    indexes: Optional[List[Union[int, str, datetime.datetime]]],
) -> Optional[List[Union[int, str, datetime.datetime]]]:
    try:
        if indexes and type(indexes[0]) is str and not str(indexes[0]).isnumeric():
            return list(map(lambda dt: pd.Timestamp(dt), indexes))
        if indexes and type(indexes[0]) is str and str(indexes[0]).isnumeric():
            return list(map(lambda idx: int(str(idx)), indexes))
        return indexes
    except ValueError as e:
        raise ValueError("Could not parse indexes as either int or dates", e)


if TYPE_CHECKING:
    from .ensemble_model import EnsembleModel
    from webviz_ert.plugins import WebvizErtPluginABC


def load_ensemble(
    parent_page: "WebvizErtPluginABC", ensemble_id: str
) -> "EnsembleModel":
    ensemble = parent_page.get_ensemble(ensemble_id=ensemble_id)
    if ensemble is None:
        ensemble = EnsembleModel(
            ensemble_id=ensemble_id, project_id=parent_page.project_identifier
        )
        parent_page.add_ensemble(ensemble)
    return ensemble


from .data_model import DataType, AxisType
from .observation import Observation
from .realization import Realization
from .response import Response
from .plot_model import (
    PlotModel,
    ResponsePlotModel,
    HistogramPlotModel,
    MultiHistogramPlotModel,
    BoxPlotModel,
    ParallelCoordinatesPlotModel,
    BarChartPlotModel,
)
from .parameter_model import PriorModel, ParametersModel
from .ensemble_model import EnsembleModel



import json
import pandas as pd
from typing import Mapping, List, Dict, Union, Any, Optional
from webviz_ert.data_loader import get_data_loader, DataLoaderException
from webviz_ert.models import Response, PriorModel, ParametersModel


def _create_parameter_models(
    parameters_names: list,
    priors: dict,
    ensemble_id: str,
    project_id: str,
) -> Optional[Mapping[str, ParametersModel]]:
    parameters = {}
    for param in parameters_names:
        key = param
        prior_schema = priors.get(key, None)
        prior = None
        if prior_schema:
            prior = PriorModel(
                prior_schema["function"],
                [x[0] for x in prior_schema.items() if isinstance(x[1], (float, int))],
                [x[1] for x in prior_schema.items() if isinstance(x[1], (float, int))],
            )

        parameters[key] = ParametersModel(
            group="",  # TODO?
            key=key,
            prior=prior,
            param_id="",  # TODO?
            project_id=project_id,
            ensemble_id=ensemble_id,
        )
    return parameters


class EnsembleModel:
    def __init__(self, ensemble_id: str, project_id: str) -> None:
        self._data_loader = get_data_loader(project_id)
        self._schema = self._data_loader.get_ensemble(ensemble_id)
        self._experiment_id = self._schema["experiment_id"]
        self._project_id = project_id
        self._metadata = self._schema["userdata"]
        self._name = self._metadata["name"]
        self._id = ensemble_id
        self._children = self._schema.get("child_ensemble_ids", [])
        self._parent = self._schema.get("parent_ensemble_id", [])
        self._size = self._schema["size"]
        self._active_realizations = self._schema["active_realizations"]
        self._responses: Dict[str, Response] = {}
        self._parameters: Optional[Mapping[str, ParametersModel]] = None
        self._cached_children: Optional[List["EnsembleModel"]] = None
        self._cached_parent: Optional["EnsembleModel"] = None

    @property
    def responses(self) -> Dict[str, Response]:
        if not self._responses:
            self._responses = {}
            responses_dict = self._data_loader.get_ensemble_responses(self._id)
            response_names = list(responses_dict.keys())
            response_names.sort()

            for name in response_names:
                self._responses[name] = Response(
                    name=name,
                    ensemble_id=self._id,
                    project_id=self._project_id,
                    ensemble_size=self._size,
                    active_realizations=self._active_realizations,
                    resp_schema=responses_dict[name],
                )
        return self._responses

    @property
    def children(self) -> Optional[List["EnsembleModel"]]:
        if not self._cached_children:
            self._cached_children = [
                EnsembleModel(
                    ensemble_id=child,
                    project_id=self._project_id,
                )
                for child in self._children
            ]
        return self._cached_children

    @property
    def parent(self) -> Optional["EnsembleModel"]:
        if not self._parent:
            return None
        if not self._cached_parent:
            self._cached_parent = EnsembleModel(
                ensemble_id=self._parent,
                project_id=self._project_id,
            )
        return self._cached_parent

    @property
    def parameters(
        self,
    ) -> Optional[Mapping[str, ParametersModel]]:
        if not self._parameters:
            parameter_names = []
            for params in self._data_loader.get_ensemble_parameters(self._id):
                labels = params["labels"]
                param_name = params["name"]
                if len(labels) > 0:
                    for label in labels:
                        parameter_names.append(f"{param_name}::{label}")
                else:
                    parameter_names.append(param_name)
            parameter_priors = (
                self._data_loader.get_experiment_priors(self._experiment_id)
                if not self._parent
                else {}
            )
            self._parameters = _create_parameter_models(
                parameter_names,
                parameter_priors,
                ensemble_id=self._id,
                project_id=self._project_id,
            )
        return self._parameters

    def parameters_df(self, parameter_list: Optional[List[str]] = None) -> pd.DataFrame:
        if not self.parameters or not parameter_list:
            return pd.DataFrame()
        data = {
            parameter: self.parameters[parameter].data_df().values.flatten()
            for parameter in parameter_list
        }
        return pd.DataFrame(data=data)

    @property
    def id(self) -> str:
        return self._id

    def __str__(self) -> str:
        return f"{self._name}"

    def __repr__(self) -> str:
        return f"{self.id}, {self._name}"

    @property
    def name(self) -> str:
        return self._name



import pytest
from webviz_ert.plugins import ObservationAnalyzer
from tests.conftest import (
    get_options,
    select_ensemble,
    verify_key_in_dropdown,
    setup_plugin,
)


@pytest.mark.browser_test
def test_observation_analyzer_view_ensemble_no_observations(
    mock_data,
    dash_duo,
):
    # This test selects an ensemble from the ensemble-multi-selector
    # where no response has any observations, and checks that no responses are
    # available in the response selector
    plugin = setup_plugin(dash_duo, __name__, ObservationAnalyzer)

    ensemble_name = select_ensemble(dash_duo, plugin)
    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("selected-ensemble-dropdown"),
        ensemble_name,
    )

    response_options_selector = f"#{plugin.uuid('response-selector')}"
    dash_duo.wait_for_text_to_equal(response_options_selector, "Select...")

    # assert dash_duo.get_logs() == [], "browser console should contain no error"


@pytest.mark.browser_test
def test_observation_analyzer_view_ensemble_with_observations(
    mock_data,
    dash_duo,
):
    # This test selects an ensemble from the ensemble-multi-selector that has
    # responses with observations, and checks that a response with observations
    # is present among the choosable responses.
    plugin = setup_plugin(dash_duo, __name__, ObservationAnalyzer)

    ensemble_name = select_ensemble(dash_duo, plugin, "default3")

    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("selected-ensemble-dropdown"),
        ensemble_name,
    )

    verify_key_in_dropdown(dash_duo, plugin.uuid("response-selector"), "FOPR")

    # assert dash_duo.get_logs() == [], "browser console should contain no error"



import logging
import dash
import pandas as pd
import datetime
import plotly.graph_objects as go
import webviz_ert.assets as assets

from copy import deepcopy
from typing import List, Dict, Union, Any, Optional, Type, Tuple, TYPE_CHECKING
from dash.dependencies import Input, Output, State, ALL, MATCH
from dash.exceptions import PreventUpdate
from webviz_ert.plugins import WebvizErtPluginABC
from webviz_ert.models import (
    ResponsePlotModel,
    Response,
    PlotModel,
    load_ensemble,
    AxisType,
)

if TYPE_CHECKING:
    from .ensemble_model import EnsembleModel

logger = logging.getLogger()


def _get_realizations_plots(
    realizations_df: pd.DataFrame,
    x_axis: Optional[List[Union[int, str, datetime.datetime]]],
    color: str,
    style: Optional[Dict] = None,
    ensemble_name: str = "",
) -> List[PlotModel]:
    if style:
        _style = style
    else:
        _style = assets.ERTSTYLE["response-plot"]["response"].copy()
    _style.update({"line": {"color": color}})
    _style.update({"marker": {"color": color}})
    realizations_data = list()
    for idx, realization in enumerate(realizations_df):
        plot = PlotModel(
            x_axis=x_axis,
            y_axis=realizations_df[realization].values,
            text=f"Realization: {realization} Ensemble: {ensemble_name}",
            name=ensemble_name,
            legendgroup=ensemble_name,
            showlegend=False if idx > 0 else True,
            **_style,
        )
        realizations_data.append(plot)
    return realizations_data


def _get_realizations_statistics_plots(
    df_response: pd.DataFrame,
    x_axis: Optional[List[Union[int, str, datetime.datetime]]],
    color: str,
    ensemble_name: str = "",
) -> List[PlotModel]:
    data = df_response
    p10 = data.quantile(0.1, axis=1)
    p90 = data.quantile(0.9, axis=1)
    _mean = data.mean(axis=1)
    style = deepcopy(assets.ERTSTYLE["response-plot"]["statistics"])
    style["line"]["color"] = color
    style_mean = deepcopy(style)
    style_mean["line"]["dash"] = "solid"
    mean_data = PlotModel(
        x_axis=x_axis,
        y_axis=_mean,
        text=f"Mean",
        name=f"Mean {ensemble_name}",
        **style_mean,
    )
    lower_std_data = PlotModel(
        x_axis=x_axis, y_axis=p10, text="p10 quantile", name="p10 quantile", **style
    )
    upper_std_data = PlotModel(
        x_axis=x_axis, y_axis=p90, text="p90 quantile", name="p90 quantile", **style
    )
    return [mean_data, lower_std_data, upper_std_data]


def _get_observation_plots(
    observation_df: pd.DataFrame,
    metadata: Optional[List[str]] = None,
    ensemble: str = "",
) -> PlotModel:
    data = observation_df["values"]
    stds = observation_df["std"]
    x_axis = observation_df["x_axis"]
    attributes = observation_df["attributes"]
    active_mask = observation_df["active"]

    style = deepcopy(assets.ERTSTYLE["response-plot"]["observation"])
    color = [style["color"] if active else "rgb(0, 0, 0)" for active in active_mask]
    style["marker"]["color"] = color

    observation_data = PlotModel(
        x_axis=x_axis,
        y_axis=data,
        text=attributes,
        name=f"Observation_{ensemble}",
        error_y=dict(
            type="data",  # value of error bar given in data coordinates
            array=stds.values,
            visible=True,
        ),
        meta=metadata,
        **style,
    )
    return observation_data


def _create_response_plot(
    response: Response,
    plot_type: str,
    selected_realizations: List[int],
    color: str,
    style: Optional[Dict] = None,
    ensemble_name: str = "",
) -> ResponsePlotModel:
    x_axis = response.axis
    if plot_type == "Statistics":
        realizations = _get_realizations_statistics_plots(
            response.data_df(selected_realizations),
            x_axis,
            color=color,
            ensemble_name=ensemble_name,
        )
    else:
        realizations = _get_realizations_plots(
            response.data_df(selected_realizations),
            x_axis,
            color=color,
            style=style,
            ensemble_name=ensemble_name,
        )
    if response.observations:
        observations = [
            _get_observation_plots(obs.data_df(), ensemble=ensemble_name)
            for obs in response.observations
        ]
    else:
        observations = []

    ensemble_plot = ResponsePlotModel(realizations, observations, dict())
    return ensemble_plot


def multi_response_controller(parent: WebvizErtPluginABC, app: dash.Dash) -> None:
    @app.callback(
        Output(
            {
                "index": MATCH,
                "id": parent.uuid("response-graphic"),
                "type": parent.uuid("graph"),
            },
            "figure",
        ),
        [
            Input({"index": MATCH, "type": parent.uuid("plot-type")}, "value"),
            Input(parent.uuid("ensemble-selection-store"), "modified_timestamp"),
        ],
        [
            State(parent.uuid("selected-ensemble-dropdown"), "value"),
            State({"index": MATCH, "type": parent.uuid("response-id-store")}, "data"),
        ],
    )
    def update_graph(
        plot_type: str,
        _: Any,
        selected_ensembles: List[str],
        response: Optional[str],
    ) -> go.Figure:
        if not response or not selected_ensembles:
            raise PreventUpdate

        def _generate_plot(
            ensemble: "EnsembleModel", color: str
        ) -> Optional[ResponsePlotModel]:
            if response not in ensemble.responses:
                return None
            plot = _create_response_plot(
                ensemble.responses[response],
                plot_type,
                [],
                color,
                ensemble_name=ensemble.name,
            )
            return plot

        loaded_ensembles = [
            load_ensemble(parent, ensemble_id) for ensemble_id in selected_ensembles
        ]

        response_plots = [
            _generate_plot(ensemble, assets.get_color(index=index))
            for index, ensemble in enumerate(loaded_ensembles)
        ]

        x_axis_label = axis_label_for_ensemble_response(loaded_ensembles[0], response)

        fig = go.Figure()
        for plot in filter(None, response_plots):
            for realization in plot._realization_plots:
                fig.add_trace(realization.repr)
        for plot in filter(None, response_plots):
            for observation in plot._observations:
                fig.add_trace(observation.repr)
        fig.update_layout(assets.ERTSTYLE["figure"]["layout"])
        fig.update_layout({"xaxis": {"title": {"text": x_axis_label}}})
        fig.update_layout(assets.ERTSTYLE["figure"]["layout-value-y-axis-label"])
        return fig


def axis_label_for_ensemble_response(
    ensemble: "EnsembleModel", response_name: str
) -> str:
    response: Response = ensemble.responses[response_name]
    if response.axis_type == AxisType.TIMESTAMP:
        return "Date"
    return "Index"



from dash.development.base_component import Component
from webviz_ert.plugins import WebvizErtPluginABC

from dash import html
from dash import dcc
import webviz_core_components as wcc
import dash_bootstrap_components as dbc

from webviz_ert.models.data_model import DataType


def parameter_selector_view(
    parent: WebvizErtPluginABC,
    data_type: DataType,
    titleLabel: str = "Parameters",
) -> Component:
    return html.Div(
        [
            dbc.Row(
                [
                    dbc.Col(
                        html.H6(
                            titleLabel,
                            className="ert-label",
                        ),
                        align="left",
                    ),
                    dbc.Col(
                        [
                            html.Label(
                                "Search: ",
                                className="ert-label",
                            ),
                            dcc.Input(
                                id=parent.uuid(
                                    f"parameter-selector-filter-{data_type}"
                                ),
                                type="search",
                                placeholder="Substring...",
                                persistence="session",
                            ),
                        ],
                        align="right",
                        width="auto",
                    ),
                ],
                align="center",
            ),
            html.Div(
                wcc.Select(
                    id=parent.uuid(f"parameter-selector-multi-{data_type}"),
                    multi=True,
                    size=10,
                    persistence="session",
                    className="ert-select-variable",
                ),
                id=parent.uuid(f"container-parameter-selector-multi-{data_type}"),
                className="ert-parameter-selector-container-show",
            ),
            dcc.Dropdown(
                id=parent.uuid(f"parameter-deactivator-{data_type}"),
                multi=True,
                searchable=False,
                placeholder="",
                persistence="session",
            ),
            dcc.Store(
                id=parent.uuid(f"parameter-selection-store-{data_type}"),
                storage_type="session",
                data=parent.load_state(f"{data_type}", []),
            ),
        ],
    )



from typing import List
from dash.development.base_component import Component
from webviz_ert.plugins import WebvizErtPluginABC


from dash import html
from dash import dcc


def response_obs_view(parent: WebvizErtPluginABC) -> List[Component]:
    return [
        html.H5("Observation/Misfits plots"),
        html.Div(
            className="ert-dropdown-container",
            children=[
                html.Label("Response", className="ert-label"),
                dcc.Dropdown(
                    id=parent.uuid("response-selector"),
                    className="ert-dropdown",
                ),
                dcc.Store(
                    id=parent.uuid("response-selector-store"),
                    data=parent.load_state("response"),
                    storage_type="session",
                ),
            ],
        ),
        html.Div(
            [
                html.Div(
                    className="ert-graph-options",
                    children=[
                        html.Label("Y-axis type:"),
                        dcc.RadioItems(
                            options=[
                                {"label": key, "value": key}
                                for key in ["linear", "log"]
                            ],
                            value="linear",
                            id=parent.uuid("yaxis-type"),
                        ),
                        html.Label("Misfits Type:"),
                        dcc.RadioItems(
                            options=[
                                {"label": key, "value": key}
                                for key in ["Univariate", "Summary"]
                            ],
                            value="Univariate",
                            id=parent.uuid("misfits-type"),
                        ),
                    ],
                ),
                dcc.Graph(
                    id={
                        "id": parent.uuid("response-graphic"),
                        "type": parent.uuid("graph"),
                    },
                    className="ert-graph",
                ),
            ],
            className="ert-graph-container",
            id=parent.uuid("observations-graph-container"),
        ),
    ]



from typing import List
from typing import List
from dash.development.base_component import Component
from webviz_config import WebvizPluginABC

from dash import html
from dash import dcc
import dash_bootstrap_components as dbc
import webviz_ert.assets as assets


def parameter_view(parent: WebvizPluginABC, index: str = "") -> List[Component]:
    return [
        dcc.Store(
            id={"index": index, "type": parent.uuid("parameter-id-store")}, data=index
        ),
        dbc.Row(
            children=[
                dbc.Col(
                    [
                        html.H4(
                            index,
                            className="ert-parameter-view-caption",
                            id={
                                "index-caption": index,
                                "type": parent.uuid("parameter-id-store"),
                            },
                        )
                    ],
                    align="left",
                ),
                dbc.Tooltip(
                    index,
                    target={
                        "index-caption": index,
                        "type": parent.uuid("parameter-id-store"),
                    },
                    placement="bottom-start",
                    class_name="ert-parameter-view-caption-tooltip",
                ),
            ]
        ),
        dbc.Row(
            children=[
                dbc.Col(
                    [
                        html.Label("Plots:"),
                    ],
                    width="auto",
                    align="left",
                ),
                dbc.Col(
                    [
                        dcc.Checklist(
                            id={
                                "index": index,
                                "type": parent.uuid("hist-check"),
                            },
                            options=[
                                {"label": "histogram", "value": "hist"},
                                {"label": "kde", "value": "kde"},
                            ],
                            value=["hist", "kde"],
                            persistence="session",
                        ),
                    ],
                    align="left",
                    width="auto",
                ),
                dbc.Col(
                    [
                        html.Label("Number of bins:", className="ert-label"),
                    ],
                    align="left",
                    width="auto",
                ),
                dbc.Col(
                    [
                        dcc.Input(
                            id={
                                "index": index,
                                "type": parent.uuid("hist-bincount"),
                            },
                            type="number",
                            placeholder="# bins",
                            min=2,
                            debounce=True,
                            className="ert-input-number",
                        ),
                    ],
                    align="left",
                    width="auto",
                ),
                dcc.Store(
                    id={"index": index, "type": parent.uuid("bincount-store")},
                    storage_type="session",
                ),
            ],
        ),
        dcc.Graph(
            id={
                "index": index,
                "id": parent.uuid("parameter-scatter"),
                "type": parent.uuid("graph"),
            },
            config={"responsive": True},
            style={"height": "450px"},
        ),
    ]



import pytest
from requests import HTTPError
import dash
from selenium.webdriver.chrome.options import Options
from selenium.common.exceptions import TimeoutException

from tests.data.snake_oil_data import ensembles_response


def pytest_addoption(parser):
    parser.addoption(
        "--skip-browser-tests",
        action="store_true",
        default=False,
        help="This option allows skipping tests that depend on chromedriver",
    )


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "browser_test: mark test as chromedriver dependent"
    )


def pytest_collection_modifyitems(config, items):
    skip_browser_tests = pytest.mark.skip(
        reason="chromedriver missing in PATH or intentionally skipped"
    )
    browser_tests = [item for item in items if "browser_test" in item.keywords]
    if config.getoption("--skip-browser-tests"):
        for item in browser_tests:
            item.add_marker(skip_browser_tests)


def pytest_setup_options():
    options = Options()
    options.add_argument("--headless")
    options.add_argument("--no-sandbox")
    options.add_argument("--disable-dev-shm-usage")
    options.add_argument("--disable-setuid-sandbox")
    return options


@pytest.fixture
def mock_data(mocker):
    mocker.patch(
        "webviz_ert.data_loader.get_connection_info",
        side_effect=lambda _: {"baseurl": "http://127.0.0.1:5000", "auth": ""},
    )

    mocker.patch("webviz_ert.data_loader._requests_get", side_effect=_requests_get)
    mocker.patch("webviz_ert.data_loader._requests_post", side_effect=_requests_post)


class _MockResponse:
    def __init__(self, url, data, status_code):
        self.url = url
        self.data = data
        self.status_code = status_code

    def json(self):
        return self.data

    @property
    def text(self):
        return self.data

    @property
    def raw(self):
        return self.data

    @property
    def content(self):
        return self.data

    def raise_for_status(self):
        if self.status_code == 400:
            raise HTTPError(
                "Mocked requests raised HTTPError 400 due to missing data in "
                "test-data set!\n"
                f"{self.url}"
            )


def _requests_get(url, **kwargs):
    if kwargs.get("params") is not None:
        url += "?"
        for param, value in kwargs["params"].items():
            url += f"{param}={value}"
    if url in ensembles_response:
        return _MockResponse(url, ensembles_response[url], 200)
    return _MockResponse(url, {}, 400)


def _requests_post(url, **kwargs):
    if url in ensembles_response:
        return _MockResponse(url, ensembles_response[url], 200)
    return _MockResponse(url, {}, 400)


def select_first(dash_duo, selector):
    options = dash_duo.find_elements(selector + " option")
    if not options:
        raise AssertionError(f"No selection option for selector {selector}")
    text = options[0].text
    options[0].click()
    # TODO remove wait?
    wait_a_bit(dash_duo, time_seconds=0.5)
    return text


def select_by_name(dash_duo, selector, name):
    dash_duo.wait_for_contains_text(selector, name)
    options = dash_duo.find_elements(selector + " option")
    if not options:
        raise AssertionError(f"No `option`s under selector {selector}")
    for option in options:
        if option.text == name:
            option.click()
            # TODO remove wait?
            wait_a_bit(dash_duo, time_seconds=0.5)
            return name
    raise AssertionError(f"Option {name} not available in {selector}")


def get_options(dash_duo, selector):
    parameter_selector_input = dash_duo.find_element(selector)
    return parameter_selector_input.text.split("\n")


def setup_plugin(
    dash_duo,
    name,
    plugin_class,
    window_size=(630, 2000),
    project_identifier=None,
    beta: bool = False,
):
    app = dash.Dash(name)
    plugin = plugin_class(app, project_identifier=project_identifier, beta=beta)
    app.layout = plugin.layout
    dash_duo.start_server(app)
    windowsize = window_size
    dash_duo.driver.set_window_size(*windowsize)
    return plugin


def select_ensemble(dash_duo, plugin, wanted_ensemble_name=None):
    """tries to select, i.e. click, the ensemble given by ensemble_name, and
    clicks the first one if no name is given. It returns the name of the
    selected ensemble."""
    ensemble_selector_id = plugin.uuid("ensemble-multi-selector")
    if not wanted_ensemble_name:
        first_ensemble_name = select_first(dash_duo, "#" + ensemble_selector_id)
        dash_duo.wait_for_contains_text(
            "#" + plugin.uuid("selected-ensemble-dropdown"),
            first_ensemble_name,
            timeout=4,
        )
        return first_ensemble_name

    options_selector_prefix = f"#{ensemble_selector_id}"
    select_by_name(dash_duo, options_selector_prefix, wanted_ensemble_name)
    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("selected-ensemble-dropdown"),
        wanted_ensemble_name,
        timeout=4,
    )
    return wanted_ensemble_name


def select_response(dash_duo, plugin, response_name=None, wait_for_plot=True) -> str:
    response_selector_id = f'#{plugin.uuid("parameter-selector-multi-resp")}'
    if response_name is None:
        response_name = select_first(dash_duo, response_selector_id)
    else:
        select_by_name(
            dash_duo,
            response_selector_id,
            response_name,
        )
    if wait_for_plot:
        dash_duo.wait_for_element(f'[id="{plugin.uuid(response_name)}"]')
    return response_name


def select_parameter(dash_duo, plugin, parameter_name=None, wait_for_plot=True) -> str:
    parameter_selector_id = f'#{plugin.uuid("parameter-selector-multi-param")}'
    if parameter_name is None:
        parameter_name = select_first(dash_duo, parameter_selector_id)
    else:
        select_by_name(
            dash_duo,
            parameter_selector_id,
            parameter_name,
        )
    if wait_for_plot:
        dash_duo.wait_for_element(f'[id="{plugin.uuid(parameter_name)}"]')
    return parameter_name


def wait_a_bit(dash_duo, time_seconds=0.1):
    try:
        dash_duo.wait_for_element(".foo-elderberries-baarrrrr", timeout=time_seconds)
    except TimeoutException:
        pass


def verify_key_in_dropdown(dash_duo, selector, key):
    verify_keys_in_dropdown(dash_duo, selector, [key])


def verify_keys_in_dropdown(dash_duo, selector, keys):
    dropdown = dash_duo.find_element(f"#{selector}")
    dropdown.click()
    for key in keys:
        dash_duo.wait_for_contains_text(f"#{selector} div.Select-menu-outer", key)



from typing import List, Optional
import pandas as pd
import math


class Realization:
    def __init__(self, realization_schema: dict):
        self._name = realization_schema["name"]
        self._data = realization_schema["data"]
        self._univariate_misfits_df = self._extract_univariate_misfits(
            realization_schema.get("univariate_misfits")
        )

    def _extract_univariate_misfits(
        self, schema: Optional[dict]
    ) -> Optional[pd.DataFrame]:
        if not schema:
            return None

        misfits_list = []
        for _, misfits in schema.items():
            for misfits_instance in misfits:
                misfits_list.append(misfits_instance)

        df = pd.DataFrame(data=misfits_list)
        df["value_sign"] = df[["value", "sign"]].apply(
            lambda row: -1.0 * math.sqrt(row[0]) if row[1] else math.sqrt(row[0]),
            axis=1,
        )
        return df

    @property
    def summarized_misfits_value(self) -> float:
        return self.univariate_misfits_df["value"].sum()

    @property
    def univariate_misfits_df(self) -> pd.DataFrame:
        return self._univariate_misfits_df

    @property
    def data(self) -> List[float]:
        return self._data

    @property
    def name(self) -> str:
        return self._name



import io
import json
import logging
import operator
from collections import defaultdict
from typing import Any, Callable, List, Mapping, MutableMapping, Optional, Tuple
from urllib.parse import quote
from uuid import UUID

import pandas as pd
import requests

logger = logging.getLogger()

connection_info_map: dict = {}


def escape(s: str) -> str:
    return quote(quote(s, safe=""))


def get_connection_info(project_id: str) -> Mapping[str, str]:
    from ert.shared.storage.connection import get_info

    if project_id not in connection_info_map:
        info = get_info(project_id)
        info["auth"] = info["auth"][1]
        connection_info_map[project_id] = info

    return connection_info_map[project_id]


# these are needed to mock for testing
def _requests_get(*args: Any, **kwargs: Any) -> requests.models.Response:
    return requests.get(*args, **kwargs)


def _requests_post(*args: Any, **kwargs: Any) -> requests.models.Response:
    return requests.post(*args, **kwargs)


data_cache: dict = {}
ServerIdentifier = Tuple[str, Optional[str]]  # (baseurl, optional token)


class DataLoaderException(Exception):
    pass


class DataLoader:
    _instances: MutableMapping[ServerIdentifier, "DataLoader"] = {}

    baseurl: str
    token: Optional[str]
    cert: Optional[str]
    _graphql_cache: MutableMapping[str, MutableMapping[dict, Any]]

    def __new__(
        cls, baseurl: str, token: Optional[str] = None, cert: Optional[str] = None
    ) -> "DataLoader":
        if (baseurl, token) in cls._instances:
            return cls._instances[(baseurl, token)]

        loader = super().__new__(cls)
        loader.baseurl = baseurl
        loader.token = token
        loader.cert = cert
        loader._graphql_cache = defaultdict(dict)
        cls._instances[(baseurl, token)] = loader
        return loader

    def _get(
        self, url: str, headers: Optional[dict] = None, params: Optional[dict] = None
    ) -> requests.Response:
        if headers is None:
            headers = {}

        resp = _requests_get(
            f"{self.baseurl}/{url}",
            headers={**headers, "Token": self.token},
            verify=self.cert,
            params=params,
        )
        if resp.status_code != 200:
            raise DataLoaderException(
                f"""Error fetching data from {self.baseurl}/{url}
                The request return with status code: {resp.status_code}
                {str(resp.content)}
                """
            )
        return resp

    def _post(
        self, url: str, headers: Optional[dict] = None, params: Optional[dict] = None
    ) -> requests.Response:
        if headers is None:
            headers = {}

        resp = _requests_post(
            f"{self.baseurl}/{url}",
            headers={**headers, "Token": self.token},
            params=params,
        )
        if resp.status_code != 200:
            raise DataLoaderException(
                f"""Error posting to {self.baseurl}/{url}
                The request returned status code: {resp.status_code}
                {str(resp.content)}
                """
            )
        return resp

    def get_all_ensembles(self) -> list:
        try:
            experiments = self._get(url=f"experiments").json()
            return [
                self.get_ensemble(ensemble_id)
                for experiment in experiments
                for ensemble_id in experiment["ensemble_ids"]
            ]
        except DataLoaderException as e:
            logger.error(e)
            return list()

    def get_ensemble(self, ensemble_id: str) -> dict:
        try:
            return self._get(url=f"ensembles/{ensemble_id}").json()
        except DataLoaderException as e:
            logger.error(e)
            return dict()

    def get_ensemble_responses(self, ensemble_id: str) -> dict:
        try:
            # Map to experiment ID
            ensemble_info = self.get_ensemble(ensemble_id)
            experiment_id = ensemble_info["experiment_id"]
            experiment_info = self._get(url=f"experiments/{experiment_id}").json()

            # Read response/ obs metadata
            responses = experiment_info["responses"]
            observations = experiment_info["observations"]

            # Expand gendatas to the @ notation
            response_map = {}
            for gendata in responses.get("gen_data", []):
                report_steps = gendata["filter_on"]["report_step"]
                response_key = gendata["response_key"]
                for report_step in report_steps:
                    legacy_gendata_key = f"{response_key}@{report_step}"
                    response_map[legacy_gendata_key] = {
                        "id": f"id={UUID(int=0)}",
                        "name": legacy_gendata_key,
                        "has_observations": bool(
                            observations.get("gen_data", {}).get(response_key)
                        ),
                        "userdata": {"data_origin": "GEN_DATA"},
                    }

            for smry in responses.get("summary", []):
                response_key = smry["response_key"]
                response_map[response_key] = {
                    "id": f"id={UUID(int=0)}",
                    "name": response_key,
                    "has_observations": bool(
                        observations.get("summary", {}).get(response_key)
                    ),
                    "userdata": {"data_origin": "Summary"},
                }

            return response_map

        except DataLoaderException as e:
            logger.error(e)
            return dict()

    def get_ensemble_userdata(self, ensemble_id: str) -> dict:
        try:
            return self._get(url=f"ensembles/{ensemble_id}/userdata").json()
        except DataLoaderException as e:
            logger.error(e)
            return dict()

    def get_ensemble_parameters(self, ensemble_id: str) -> list:
        try:
            ensemble_info = self.get_ensemble(ensemble_id)
            experiment_id = ensemble_info["experiment_id"]
            experiment_info = self._get(url=f"experiments/{experiment_id}").json()

            # Read response/ obs metadata
            parameters = experiment_info["parameters"]

            param_list = []
            for param_name, metadata in parameters.items():
                userdata = {"data_origin": metadata["type"]}
                is_log = False
                if distribution := metadata.get("distribution"):
                    is_log = distribution["name"] in {
                        "lognormal",
                        "logunif",
                    }
                param_list.append(
                    {
                        "userdata": userdata,
                        "dimensionality": metadata["dimensionality"],
                        "name": (f"LOG10_{param_name}" if is_log else f"{param_name}"),
                        "labels": [],
                    }
                )

            return param_list
        except DataLoaderException as e:
            logger.error(e)
            return list()

    def get_experiment_priors(self, experiment_id: str) -> dict:
        try:
            experiment = self._get(url=f"experiments/{experiment_id}").json()
            return experiment["priors"]
        except RuntimeError as e:
            logger.error(e)
            return dict()

    def get_ensemble_parameter_data(
        self,
        ensemble_id: str,
        parameter_name: str,
    ) -> pd.DataFrame:
        try:
            if "::" in parameter_name:
                name, label = parameter_name.split("::", 1)
                params = {"label": label}
            else:
                name = parameter_name
                params = {}

            resp = self._get(
                url=f"ensembles/{ensemble_id}/parameters/{escape(name)}",
                headers={"accept": "application/x-parquet"},
                params=params,
            )
            stream = io.BytesIO(resp.content)
            df = pd.read_parquet(stream).transpose()
            return df
        except DataLoaderException as e:
            logger.error(e)
            return pd.DataFrame()

    def get_ensemble_record_data(
        self,
        ensemble_id: str,
        record_name: str,
    ) -> pd.DataFrame:
        try:
            if "@" in record_name:
                response_key, report_step = record_name.split("@", 1)
                resp = self._get(
                    url=f"ensembles/{ensemble_id}/responses/{escape(response_key)}",
                    params={"filter_on": json.dumps({"report_step": report_step})},
                    headers={"accept": "application/x-parquet"},
                )
            else:
                resp = self._get(
                    url=f"ensembles/{ensemble_id}/responses/{escape(record_name)}",
                    headers={"accept": "application/x-parquet"},
                )

            stream = io.BytesIO(resp.content)
            df = pd.read_parquet(stream).transpose()

        except DataLoaderException as e:
            logger.error(e)
            return pd.DataFrame()
        try:
            df.index = df.index.astype(int)
        except (TypeError, ValueError):
            try:
                df.index = df.index.map(pd.Timestamp)
            except ValueError:
                pass
        df = df.sort_index()
        return df

    def get_ensemble_record_observations(
        self, ensemble_id: str, record_name: str
    ) -> List[dict]:
        try:
            if "@" in record_name:
                response_key, _ = record_name.split("@", 1)
                return self._get(
                    url=f"ensembles/{ensemble_id}/responses/{escape(response_key)}/observations",
                    # Hard coded to zero, as all realizations are connected to the same observations
                    params={"realization_index": 0},
                ).json()
            else:
                return self._get(
                    url=f"ensembles/{ensemble_id}/responses/{escape(record_name)}/observations",
                    # Hard coded to zero, as all realizations are connected to the same observations
                    params={"realization_index": 0},
                ).json()
        except DataLoaderException as e:
            logger.error(e)
            return list()

    def compute_misfit(
        self, ensemble_id: str, response_name: str, summary: bool
    ) -> pd.DataFrame:
        try:

            if not summary:
                # assume gen data
                response_key, report_step = response_name.split("@", 1)
                resp = self._get(
                    "compute/misfits",
                    params={
                        "ensemble_id": ensemble_id,
                        "response_name": response_key,
                        "filter_on": json.dumps({"report_step": report_step}),
                        "summary_misfits": False,
                    },
                )
            else:
                resp = self._get(
                    "compute/misfits",
                    params={
                        "ensemble_id": ensemble_id,
                        "response_name": response_name,
                        "summary_misfits": True,
                    },
                )
            stream = io.BytesIO(resp.content)
            df = pd.read_csv(stream, index_col=0, float_precision="round_trip")
            return df
        except DataLoaderException as e:
            logger.error(e)
            return pd.DataFrame()

    def refresh_data(self) -> requests.Response:
        return self._post("updates/facade")


def get_data_loader(project_id: str) -> DataLoader:
    return DataLoader(*(get_connection_info(project_id).values()))


def get_ensembles(project_id: str) -> list:
    return get_data_loader(project_id).get_all_ensembles()


def refresh_data(project_id: str) -> requests.Response:
    return get_data_loader(project_id).refresh_data()



import pytest

from tests.conftest import (
    select_by_name,
    select_ensemble,
    select_parameter,
    select_response,
    setup_plugin,
)
from webviz_ert.plugins._response_correlation import ResponseCorrelation


@pytest.mark.browser_test
def test_axes_labels(mock_data, dash_duo):
    """test_axis_labels loads two different plots and checks that axes are
    labelled correctly"""

    plugin = setup_plugin(
        dash_duo, __name__, ResponseCorrelation, window_size=(1024, 2048)
    )

    # find the right ensemble which has mock data prepared for this test
    wanted_ensemble_name = "default3"
    select_ensemble(dash_duo, plugin, wanted_ensemble_name)

    wanted_responses = ["WOPR:OP1", "FOPR"]

    # we only see one response at a time, so we choose and check one after the
    # other
    for wanted_response in wanted_responses:
        response_selector_id = plugin.uuid("parameter-selector-multi-resp")
        select_by_name(dash_duo, f"#{response_selector_id}", wanted_response)

        plot_id = plugin.uuid("response-overview")

        # check that y axis label spells out "Value"; test is flaky so longer
        # timeout.
        dash_duo.wait_for_text_to_equal(f"#{plot_id} text.ytitle", "Value", timeout=20)

        # check that one has date, the other has index as x axis label
        if wanted_response == "FOPR":
            dash_duo.wait_for_text_to_equal(f"#{plot_id} text.xtitle", "Date")
        else:
            dash_duo.wait_for_text_to_equal(f"#{plot_id} text.xtitle", "Index")

        # clear response selection so next selection can be displayed
        response_deactivator_id = plugin.uuid("parameter-deactivator-resp")
        selected_response_closer = dash_duo.find_element(
            f"#{response_deactivator_id} .Select-value-icon"
        )
        selected_response_closer.click()

        # wait for deselected option to reappear among available options
        dash_duo.wait_for_contains_text(f"#{response_selector_id}", wanted_response)

    # assert dash_duo.get_logs() == [], "browser console should contain no error"


@pytest.mark.browser_test
def test_show_respo_with_obs(mock_data, dash_duo):
    """Test response observation filter works as expected"""
    plugin = setup_plugin(dash_duo, __name__, ResponseCorrelation)

    select_ensemble(dash_duo, plugin, "default3")

    expected_responses_with_observations = ["FOPR", "WOPR:OP1"]

    response_selector_id = "#" + plugin.uuid("parameter-selector-multi-resp")

    dash_duo.wait_for_text_to_equal(
        response_selector_id,
        "\n".join(expected_responses_with_observations),
        timeout=20,
    )


@pytest.mark.browser_test
def test_info_text_appears_as_expected(
    mock_data,
    dash_duo,
):
    ensemble = "default3"
    response = "FOPR"
    parameter = "BPR_138_PERSISTENCE"
    index = "2010-01-10"
    plugin = setup_plugin(dash_duo, __name__, ResponseCorrelation)
    select_ensemble(dash_duo, plugin, ensemble)
    select_response(dash_duo, plugin, response, wait_for_plot=False)
    select_parameter(dash_duo, plugin, parameter, wait_for_plot=False)
    info_text_selector = f"#{plugin.uuid('info-text')}"
    expected_text = "".join(
        [f"RESPONSE: {response}", f"INDEX: {index}", f"PARAMETER: {parameter}"]
    )
    dash_duo.wait_for_text_to_equal(info_text_selector, expected_text, timeout=20)

    # assert dash_duo.get_logs() == [], "browser console should contain no error"



from .parameter_view import parameter_view
from .ensemble_selector_view import ensemble_selector_list
from .response_view import response_view
from .plot_view import plot_view_body, plot_view_header, plot_view_menu
from .misfit_view import response_obs_view
from .parallel_coordinates_view import parallel_coordinates_view
from .selector_view import parameter_selector_view
from .correlation_view import correlation_view
from .element_dropdown_view import element_dropdown_view



from webviz_ert.data_loader import get_ensembles, refresh_data


def test_get_ensembles(tmp_path, mock_data):
    ensembles = get_ensembles(project_id=tmp_path)
    assert [ensemble["id"] for ensemble in ensembles] == [1, 2, 3, 42]


def test_refresh_data(tmp_path, mock_data):
    resp = refresh_data(project_id=tmp_path)
    assert resp.status_code == 200



from typing import List
from dash.development.base_component import Component
from webviz_config import WebvizPluginABC

from dash import html
from dash import dcc
import dash_bootstrap_components as dbc


def response_view(parent: WebvizPluginABC, index: str = "") -> List[Component]:
    return [
        dcc.Store(
            id={"index": index, "type": parent.uuid("response-id-store")}, data=index
        ),
        dbc.Row(
            className="ert-plot-options",
            children=[
                dbc.Col(
                    [html.H4(index)],
                    align="center",
                ),
                dbc.Col(
                    [
                        html.Label("Graph Type:", className="ert-label"),
                    ],
                    width="auto",
                    align="center",
                ),
                dbc.Col(
                    [
                        dcc.RadioItems(
                            options=[
                                {"label": key, "value": key}
                                for key in ["Function plot", "Statistics"]
                            ],
                            value="Statistics",
                            id={"index": index, "type": parent.uuid("plot-type")},
                            persistence="session",
                        ),
                    ],
                    align="center",
                ),
            ],
        ),
        dcc.Graph(
            id={
                "index": index,
                "id": parent.uuid("response-graphic"),
                "type": parent.uuid("graph"),
            },
            config={"responsive": True},
        ),
    ]



import dash
import dash_daq
import webviz_ert.assets as assets
import webviz_ert.models
import webviz_ert.controllers
import dash_bootstrap_components as dbc
from dash.development.base_component import Component
from typing import List, Dict
from webviz_ert.views import (
    ensemble_selector_list,
    correlation_view,
    parameter_selector_view,
)
from webviz_ert.models.data_model import DataType
from webviz_ert.plugins import WebvizErtPluginABC


class ResponseCorrelation(WebvizErtPluginABC):
    def __init__(self, app: dash.Dash, project_identifier: str, beta: bool = False):
        super().__init__(app, project_identifier)
        self.set_callbacks(app)
        self.beta = beta

    @property
    def tour_steps(self) -> List[Dict[str, str]]:
        steps = [
            {
                "id": self.uuid("obs_index_selector_container"),
                "content": (
                    "Select preferred observation indexes by drawing a rectangle on the figure"
                ),
            },
            {
                "id": self.uuid("info-text"),
                "content": (
                    "The currently active response, parameter, and x_index / "
                    "timestamp"
                ),
            },
            {
                "id": self.uuid("response-overview"),
                "content": "Visualization of the currently active response",
            },
            {
                "id": self.uuid("response-scatterplot"),
                "content": (
                    "Scatterplot visualization of function values between currently active "
                    "response and currently active parameter. "
                    "It is accompanied by distribution plots for both selections. "
                ),
            },
            {
                "id": self.uuid("response-heatmap"),
                "content": (
                    "Heatmap based representation of correlation (-1, 1) among all selected responses "
                    "and parameters, for currently selected ensembles in column-wise fashion."
                    "One can select a currently active response at an observation index and parameter "
                    "by clicking directly on a heatmap. "
                ),
            },
            {
                "id": self.uuid("response-correlation"),
                "content": (
                    "Correlation BarChart for the currently active response (from heatmap) "
                    "and parameters in a descending order. "
                ),
            },
            {
                "id": self.uuid("correlation-metric"),
                "content": (
                    "Which metric to use, all views are automatically updated. "
                ),
            },
            {
                "id": self.uuid("sort-parameters"),
                "content": (
                    "Option for toggling sorting of parameters from sorting "
                    "by correlation to alphabetical; affects heatmap and "
                    "tornado plots."
                ),
            },
            {
                "id": self.uuid("hide-hover"),
                "content": (
                    "Option for toggling visibility of hover info text in " "heatmap."
                ),
            },
        ]

        return steps

    @property
    def layout(self) -> Component:
        return dash.html.Div(
            [
                dash.html.Div(
                    children=[
                        dash.html.P(
                            [
                                "This page is considered a ",
                                dash.html.B("beta"),
                                " version and could be changed or removed. You are encouraged to use it and give feedback regarding functionality and / or bugs.",
                            ],
                            className="ert-beta-warning",
                            id=self.uuid("beta-warning"),
                        )
                    ],
                    hidden=not self.beta,
                ),
                dash.dcc.Store(
                    id=self.uuid("correlation-store-xindex"),
                    data={},
                    storage_type="session",
                ),
                dash.dcc.Store(id=self.uuid("correlation-store-selected-obs"), data={}),
                dash.dcc.Store(
                    id=self.uuid("correlation-store-obs-range"),
                    data=self.load_state("correlation-store-obs-range", {}),
                    storage_type="session",
                ),
                dash.dcc.Store(
                    id=self.uuid("correlation-store-active-resp-param"),
                    data=self.load_state(
                        "active_correlation",
                        {"parameter": None, "response": None},
                    ),
                    storage_type="session",
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            id=self.uuid("ensemble-content"),
                            children=ensemble_selector_list(parent=self),
                            width=4,
                        ),
                        dbc.Col(
                            [
                                parameter_selector_view(
                                    self,
                                    data_type=DataType.RESPONSE,
                                    titleLabel="Responses",
                                ),
                                dash.dcc.Checklist(
                                    id=self.uuid("response-observations-check"),
                                    options=[
                                        {
                                            "label": "Show only responses with observations",
                                            "value": "obs",
                                        },
                                    ],
                                    value=["obs"],
                                    style={"display": "none"},
                                ),
                            ],
                            width=4,
                        ),
                        dbc.Col(
                            [
                                parameter_selector_view(
                                    self,
                                    data_type=DataType.PARAMETER,
                                    titleLabel="Parameters",
                                ),
                            ],
                            width=4,
                        ),
                    ],
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            [
                                dash.html.Span("Observation index selector"),
                                correlation_view(
                                    id_view=self.uuid("obs_index_selector")
                                ),
                            ],
                            className="active-info",
                            style={"min-height": "10px", "padding": "20px"},
                            id=self.uuid("obs_index_selector_container"),
                        )
                    ]
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            id=self.uuid("info-text"),
                            className="active-info",
                            children=[dash.html.Span("INFO")],
                        ),
                    ]
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            correlation_view(
                                id_view=self.uuid("response-overview"),
                            ),
                            style=assets.ERTSTYLE["dbc-column"],
                        ),
                        dbc.Col(
                            correlation_view(
                                id_view=self.uuid("response-scatterplot"),
                            ),
                            style=assets.ERTSTYLE["dbc-column"],
                        ),
                    ]
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            correlation_view(
                                id_view=self.uuid("response-correlation"),
                            ),
                            style=assets.ERTSTYLE["dbc-column"],
                        ),
                        dbc.Col(
                            correlation_view(
                                id_view=self.uuid("response-heatmap"),
                            ),
                            style=assets.ERTSTYLE["dbc-column"],
                        ),
                    ]
                ),
                dbc.Row(
                    children=[
                        dbc.Col(
                            dash.dcc.RadioItems(
                                id=self.uuid("correlation-metric"),
                                options=[
                                    {"label": "spearman", "value": "spearman"},
                                    {"label": "pearson", "value": "pearson"},
                                ],
                                value="pearson",
                            ),
                            className="correlation-option",
                            width="auto",
                        ),
                        dbc.Col(
                            children=[
                                dash_daq.BooleanSwitch(
                                    id=self.uuid("sort-parameters"),
                                    on=True,
                                    label="Sort parameters by correlation",
                                ),
                                dash_daq.BooleanSwitch(
                                    id=self.uuid("hide-hover"),
                                    on=False,
                                    label="Hide heatmap hover info",
                                ),
                            ],
                            className="heatmap-options",
                            width="auto",
                        ),
                    ],
                    justify="center",
                ),
            ]
        )

    def set_callbacks(self, app: dash.Dash) -> None:
        webviz_ert.controllers.ensemble_list_selector_controller(self, app)
        webviz_ert.controllers.parameter_selector_controller(
            self, app, data_type=DataType.PARAMETER, union_keys=False
        )
        webviz_ert.controllers.parameter_selector_controller(
            self, app, data_type=DataType.RESPONSE, extra_input=True
        )
        webviz_ert.controllers.response_correlation_controller(self, app)



import pytest
from webviz_ert.plugins import (
    ParameterComparison,
    ResponseComparison,
    ResponseCorrelation,
    ObservationAnalyzer,
)
from tests.conftest import (
    select_first,
    select_ensemble,
    select_parameter,
    select_response,
    setup_plugin,
    verify_key_in_dropdown,
)

parameter_keys = ["FIELD_PROPERTIES:POROSITY", "FIELD_PROPERTIES:X_MID_PERMEABILITY"]
response_keys = ["WGPT:PROD", "WWPT:PROD", "WOPT:PROD", "WWIT:INJ"]
response_keys_with_observations = ["WOPT:PROD"]


def _verify_keys_in_menu(dash_duo_handle, plugin, keys, selector):
    dash_duo_handle.wait_for_element("#" + plugin.uuid(selector))
    for key in keys:
        dash_duo_handle.wait_for_contains_text(
            "#" + plugin.uuid(selector),
            key,
        )


@pytest.mark.spe1
@pytest.mark.xfail(reason="Fails because ert>5 no longer supports ert-storage")
def test_webviz_parameter_comparison(dash_duo):
    plugin = setup_plugin(dash_duo, __name__, ParameterComparison)

    # Wait for the ensemble selector to be initialized
    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("ensemble-multi-selector"),
        "default",
    )

    ensemble_name = select_first(dash_duo, "#" + plugin.uuid("ensemble-multi-selector"))
    _verify_keys_in_menu(
        dash_duo, plugin, parameter_keys, "parameter-selector-multi-param"
    )


@pytest.mark.spe1
@pytest.mark.xfail(reason="Fails because ert>5 no longer supports ert-storage")
def test_webviz_response_correlation(dash_duo):
    plugin = setup_plugin(dash_duo, __name__, ResponseCorrelation)

    ensemble = "default"
    response = "WOPT:PROD"
    parameter = "FIELD_PROPERTIES:POROSITY::0"
    index = "2016-01-01"

    # Wait for the ensemble selector to be initialized
    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("ensemble-multi-selector"),
        ensemble,
    )
    select_ensemble(dash_duo, plugin, ensemble)

    _verify_keys_in_menu(
        dash_duo, plugin, parameter_keys, "parameter-selector-multi-param"
    )
    _verify_keys_in_menu(
        dash_duo,
        plugin,
        response_keys_with_observations,
        "parameter-selector-multi-resp",
    )

    select_response(dash_duo, plugin, response, wait_for_plot=False)
    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("parameter-deactivator-resp"),
        f"×{response}",
    )

    select_parameter(dash_duo, plugin, parameter, wait_for_plot=False)
    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("parameter-deactivator-param"),
        f"×{parameter}",
    )


@pytest.mark.spe1
@pytest.mark.xfail(reason="Fails because ert>5 no longer supports ert-storage")
def test_webviz_response_comparison(dash_duo):
    plugin = setup_plugin(dash_duo, __name__, ResponseComparison)

    # Wait for the ensemble selector to be initialized
    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("ensemble-multi-selector"),
        "default",
    )

    ensemble_name = select_first(dash_duo, "#" + plugin.uuid("ensemble-multi-selector"))
    _verify_keys_in_menu(
        dash_duo, plugin, parameter_keys, "parameter-selector-multi-param"
    )

    _verify_keys_in_menu(
        dash_duo, plugin, response_keys, "parameter-selector-multi-resp"
    )

    response_name = select_first(
        dash_duo, "#" + plugin.uuid("parameter-selector-multi-resp")
    )
    #':' must be escaped in element ids
    resp_plot_id = "#" + plugin.uuid(response_name.replace(":", "\\:"))
    dash_duo.wait_for_element(resp_plot_id)

    param_name = select_first(
        dash_duo, "#" + plugin.uuid("parameter-selector-multi-param")
    )
    param_plot_id = "#" + plugin.uuid(param_name.replace(":", "\\:"))
    dash_duo.wait_for_element(param_plot_id)

    # assert dash_duo.get_logs() == [], "browser console should contain no error"


@pytest.mark.spe1
@pytest.mark.xfail(reason="Fails because ert>5 no longer supports ert-storage")
def test_webviz_observation_analyzer(dash_duo):
    plugin = setup_plugin(dash_duo, __name__, ObservationAnalyzer)

    # Wait for the ensemble selector to be initialized
    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("ensemble-multi-selector"),
        "default",
    )

    ensemble_name = select_first(dash_duo, "#" + plugin.uuid("ensemble-multi-selector"))
    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("selected-ensemble-dropdown"),
        ensemble_name,
    )

    verify_key_in_dropdown(dash_duo, plugin.uuid("response-selector"), "WOPT:PROD")



import dash
import dash_bootstrap_components as dbc

from dash import html
from typing import List, Union, Any, Dict, Tuple, Optional, Mapping
from dash.development.base_component import Component
from dash.dependencies import Input, Output, State

import webviz_ert.controllers
import webviz_ert.assets as assets

from webviz_ert.views import response_view, parameter_view
from webviz_ert.plugins import WebvizErtPluginABC
from webviz_ert.models import DataType


def _new_child(parent: WebvizErtPluginABC, plot: str, data_type: DataType) -> Component:
    if data_type == DataType.RESPONSE:
        p = response_view(parent=parent, index=plot)
    if data_type == DataType.PARAMETER:
        p = parameter_view(parent=parent, index=plot)

    return dbc.Col(
        html.Div(id=parent.uuid(plot), children=p),
        xl=12,
        lg=6,
        style=assets.ERTSTYLE["dbc-column-extra-high"],
        key=plot,
    )


def plot_view_controller(
    parent: WebvizErtPluginABC, app: dash.Dash, data_type: DataType
) -> None:
    if data_type == DataType.PARAMETER:
        webviz_ert.controllers.multi_parameter_controller(parent, app)
    elif data_type == DataType.RESPONSE:
        webviz_ert.controllers.multi_response_controller(parent, app)
    else:
        raise Exception(f"Unexpected data type `{data_type}`")

    @app.callback(
        Output(parent.uuid(f"plot-selection-store-{data_type}"), "data"),
        [
            Input(
                parent.uuid(f"parameter-selection-store-{data_type}"),
                "modified_timestamp",
            ),
        ],
        [
            State(parent.uuid(f"parameter-selection-store-{data_type}"), "data"),
            State(parent.uuid(f"plot-selection-store-{data_type}"), "data"),
        ],
    )
    def update_plot_selection(
        _: Any,
        selection: Optional[List[str]],
        current_plots: Optional[List[str]],
    ) -> List[str]:
        selection = [] if not selection else selection
        current_plots = [] if not current_plots else current_plots
        for plot in current_plots.copy():
            if plot not in selection:
                current_plots.remove(plot)
        for selected in selection:
            if selected not in current_plots:
                current_plots.append(selected)

        return current_plots

    @app.callback(
        Output(parent.uuid(f"plotting-content-{data_type}"), "children"),
        Input(parent.uuid(f"plot-selection-store-{data_type}"), "data"),
        State(parent.uuid(f"plotting-content-{data_type}"), "children"),
    )
    def create_grid(
        plots: List[str],
        bootstrap_col_children: Union[None, Component, List[Component]],
    ) -> List[Component]:
        if not plots:
            return []

        children: List[Component] = (
            [] if bootstrap_col_children is None else bootstrap_col_children
        )
        if type(children) is not list:
            children = [children]

        children_names = []
        for c in children:
            children_names.append(c["props"]["key"])

        if len(plots) > len(children_names):
            # Add new plots to the grid
            for plot in plots:
                if plot not in children_names:
                    children.append(_new_child(parent, plot, data_type))
        elif len(plots) < len(children_names):
            # Remove no longer selected plots from grid
            children = list(
                filter(lambda child: child["props"]["key"] in plots, children)
            )

        for c in children:
            xl_width = 6 if len(children) > 1 else 12
            if isinstance(c, dict):
                c["props"]["xl"] = xl_width
            else:
                c.xl = xl_width

        # Every change to the childred even if the there is no change
        # and the input is returned as output triggers a redrawing
        # of every plot in the graph grid. This will sap performance
        # for very large data sets or large number of plots in the grid.
        # There is no clear way to avoid this wihtout redesigning the grid.
        return children



import dash
import pytest
from webviz_ert.assets import get_color
from webviz_ert.plugins import ParameterComparison
from tests.conftest import select_first, get_options
from tests.data.snake_oil_data import all_ensemble_names
from tests.conftest import setup_plugin, select_ensemble


@pytest.mark.browser_test
def test_ensemble_refresh(
    mock_data,
    dash_duo,
):
    plugin = setup_plugin(dash_duo, __name__, ParameterComparison)

    ensemble_selector_id = f"#{plugin.uuid('ensemble-multi-selector')}"
    expected_initial_ensembles = [
        "default",
        "default_smoother_update",
        "default3",
        "nr_42",
    ]
    dash_duo.wait_for_text_to_equal(
        ensemble_selector_id, "\n".join(expected_initial_ensembles)
    )

    # Select first ensemble
    first_ensemble_name = select_ensemble(dash_duo, plugin)

    # Select a parameter
    first_parameter = select_first(
        dash_duo, "#" + plugin.uuid("parameter-selector-multi-param")
    )

    selected_parameters_finder = (
        f"#{plugin.uuid('parameter-deactivator-param')} " ".Select-value-label"
    )
    dash_duo.wait_for_contains_text(selected_parameters_finder, first_parameter)
    # for some reason, there is an empty label
    expected_selected_parameters = [first_parameter, " "]
    dash_duo.wait_for_text_to_equal(
        selected_parameters_finder, "\n".join(expected_selected_parameters)
    )

    # Select second ensemble
    second_ensemble_name = select_ensemble(dash_duo, plugin)

    # Check selected parameters are unchanged (in particular, not duplicated)
    dash_duo.wait_for_text_to_equal(
        selected_parameters_finder, "\n".join(expected_selected_parameters)
    )

    # Check only expected selectable options are available
    expected_ensembles_after_selection = list(
        filter(
            lambda ens: ens not in (first_ensemble_name, second_ensemble_name),
            all_ensemble_names,
        )
    )
    dash_duo.wait_for_text_to_equal(
        ensemble_selector_id, "\n".join(sorted(expected_ensembles_after_selection))
    )

    # Click the refresh button
    ensemble_refresh = dash_duo.find_element(
        "#" + plugin.uuid("ensemble-refresh-button")
    )
    ensemble_refresh.click()

    # Check clicking refresh button also removes the selected parameters and options
    dash_duo.wait_for_text_to_equal(
        "#" + plugin.uuid("parameter-selector-multi-param"),
        "",
    )

    dash_duo.wait_for_text_to_equal(
        "#" + plugin.uuid("parameter-deactivator-param"),
        "",
    )

    # Check after the refresh all the initial ensemble cases are available as
    # options
    dash_duo.wait_for_text_to_equal(
        ensemble_selector_id, "\n".join(expected_initial_ensembles)
    )


@pytest.mark.browser_test
def test_ensemble_color(mock_data, dash_duo):
    plugin = setup_plugin(dash_duo, __name__, ParameterComparison, (630, 1200))

    ensembles = get_options(dash_duo, "#" + plugin.uuid("ensemble-multi-selector"))

    for _ in ensembles:
        ensemble_name = select_first(
            dash_duo, "#" + plugin.uuid("ensemble-multi-selector")
        )
        dash_duo.wait_for_contains_text(
            "#" + plugin.uuid("selected-ensemble-dropdown"),
            ensemble_name,
        )

    selected_ensembles = dash_duo.find_elements(
        "#" + plugin.uuid("selected-ensemble-dropdown") + " .Select-value"
    )
    for idx, ensemble in enumerate(selected_ensembles):
        color = ensemble.value_of_css_property("background-color")
        expected_color = get_color(idx)
        assert color.replace(" ", "") == expected_color



import pytest
import dash
from tests.conftest import setup_plugin, select_ensemble

from webviz_ert.plugins import (
    ResponseComparison,
    ObservationAnalyzer,
    ParameterComparison,
    ResponseCorrelation,
)


@pytest.mark.parametrize(
    "plugin_class,input",
    [
        pytest.param(ResponseComparison, True),
        pytest.param(ResponseComparison, False),
        pytest.param(ObservationAnalyzer, True),
        pytest.param(ObservationAnalyzer, False),
        pytest.param(ParameterComparison, True),
        pytest.param(ParameterComparison, False),
        pytest.param(ResponseCorrelation, True),
        pytest.param(ResponseCorrelation, False),
    ],
)
@pytest.mark.browser_test
def test_displaying_beta_warning(plugin_class, input: bool, dash_duo):
    plugin = setup_plugin(dash_duo, __name__, plugin_class, beta=input)
    beta_warning_element = dash_duo.find_element("#" + plugin.uuid("beta-warning"))
    assert beta_warning_element.is_displayed() == input


skip_responses = "resp"
skip_parameters = "param"


@pytest.mark.parametrize(
    "plugin_class,skip",
    [
        pytest.param(ResponseComparison, [], id="ResponseComparison"),
        pytest.param(
            ObservationAnalyzer,
            [skip_responses, skip_parameters],
            id="ObservationAnalyzer",
        ),
        pytest.param(ParameterComparison, [skip_responses], id="ParameterComparison"),
        pytest.param(ResponseCorrelation, [], id="ResponseCorrelation"),
    ],
)
@pytest.mark.browser_test
def test_selectors_visibility_toggle_button(plugin_class, skip, mock_data, dash_duo):
    # we test whether the selector visibility toggle button changes class on
    # all selectors, as expected

    plugin = setup_plugin(dash_duo, __name__, plugin_class, (2048, 1536))

    visibility_toggler = dash_duo.find_element(
        "#" + plugin.uuid("parameter-selector-button")
    )
    visibility_toggler.click()

    ensemble_container_class_prefix = "ert-ensemble-selector-container"
    dash_duo.wait_for_class_to_equal(
        f'#{plugin.uuid("container-ensemble-selector-multi")}',
        f"{ensemble_container_class_prefix}-hide",
    )

    variable_container_class_prefix = "ert-parameter-selector-container"
    if skip_responses not in skip:
        dash_duo.wait_for_class_to_equal(
            f'#{plugin.uuid("container-parameter-selector-multi-resp")}',
            f"{variable_container_class_prefix}-hide",
        )

    if skip_parameters not in skip:
        dash_duo.wait_for_class_to_equal(
            f'#{plugin.uuid("container-parameter-selector-multi-param")}',
            f"{variable_container_class_prefix}-hide",
        )

    # assert dash_duo.get_logs() == [], "browser console should contain no error"


@pytest.mark.browser_test
def test_response_selector_sorting(mock_data, dash_duo):
    plugin = setup_plugin(dash_duo, __name__, ResponseComparison)
    wanted_ensemble_name = "nr_42"
    select_ensemble(dash_duo, plugin, wanted_ensemble_name)

    response_selector_container = dash_duo.find_element(
        "#" + plugin.uuid("container-parameter-selector-multi-resp")
    )
    response_list = response_selector_container.text.split("\n")

    assert response_list == ["FGPT", "FOPR", "SNAKE_OIL_GPR_DIFF@199", "WOPR:OP1"]



import dash
import plotly.graph_objects as go

from typing import List, Tuple, Any, Optional, Mapping, Dict
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State, MATCH
from webviz_ert.models import MultiHistogramPlotModel, load_ensemble
from webviz_ert.plugins import WebvizErtPluginABC
from webviz_ert import assets


def multi_parameter_controller(parent: WebvizErtPluginABC, app: dash.Dash) -> None:
    @app.callback(
        Output({"index": MATCH, "type": parent.uuid("bincount-store")}, "data"),
        [Input({"index": MATCH, "type": parent.uuid("hist-bincount")}, "value")],
        [State({"index": MATCH, "type": parent.uuid("bincount-store")}, "data")],
    )
    def update_bincount(hist_bincount: int, store_bincount: int) -> int:
        if not isinstance(hist_bincount, int):
            raise PreventUpdate
        if hist_bincount < 2:
            raise PreventUpdate
        if hist_bincount == store_bincount:
            raise PreventUpdate
        return hist_bincount

    @app.callback(
        [
            Output(
                {
                    "index": MATCH,
                    "id": parent.uuid("parameter-scatter"),
                    "type": parent.uuid("graph"),
                },
                "figure",
            ),
            Output({"index": MATCH, "type": parent.uuid("hist-bincount")}, "value"),
        ],
        [
            Input({"index": MATCH, "type": parent.uuid("hist-check")}, "value"),
            Input(
                {"index": MATCH, "type": parent.uuid("bincount-store")},
                "modified_timestamp",
            ),
            Input(parent.uuid("ensemble-selection-store"), "modified_timestamp"),
            Input(parent.uuid("param-label-check"), "value"),
        ],
        [
            State(parent.uuid("selected-ensemble-dropdown"), "value"),
            State({"index": MATCH, "type": parent.uuid("parameter-id-store")}, "data"),
            State({"index": MATCH, "type": parent.uuid("bincount-store")}, "data"),
        ],
    )
    def update_histogram(
        hist_check_values: List[str],
        _: Any,
        __: Any,
        legend: List[str],
        selected_ensembles: List[str],
        parameter: str,
        bin_count: int,
    ) -> Tuple[go.Figure, int]:
        if not selected_ensembles:
            raise PreventUpdate

        data = {}
        colors = {}
        names = {}
        priors = {}
        for index, ensemble_id in enumerate(selected_ensembles):
            ensemble = load_ensemble(parent, ensemble_id)
            if ensemble.parameters and parameter in ensemble.parameters:
                key = str(ensemble)
                parameter_model = ensemble.parameters[parameter]
                data[key] = parameter_model.data_df()
                colors[key] = assets.get_color(index)
                names[key] = key if "label" in legend else ""

                if parameter_model.priors and "prior" in hist_check_values:
                    priors[names[key]] = (parameter_model.priors, colors[key])

        parameter_plot = MultiHistogramPlotModel(
            data,
            names=names,
            colors=colors,
            hist="hist" in hist_check_values,
            kde="kde" in hist_check_values,
            priors=priors,
            bin_count=bin_count,
        )
        return parameter_plot.repr, parameter_plot.bin_count

    @app.callback(
        Output({"index": MATCH, "type": parent.uuid("hist-check")}, "options"),
        [
            Input({"index": MATCH, "type": parent.uuid("parameter-id-store")}, "data"),
        ],
        [
            State({"index": MATCH, "type": parent.uuid("hist-check")}, "options"),
            State(parent.uuid("selected-ensemble-dropdown"), "value"),
        ],
    )
    def set_parameter_from_btn(
        parameter: str,
        plotting_options: List[Mapping[str, str]],
        selected_ensembles: List[str],
    ) -> List[Mapping[str, str]]:
        if not selected_ensembles:
            raise PreventUpdate
        has_priors = False
        for ensemble_id in selected_ensembles:
            ensemble = load_ensemble(parent, ensemble_id)
            if ensemble.parameters and parameter in ensemble.parameters:
                parameter_model = ensemble.parameters[parameter]
                if parameter_model.priors:
                    has_priors = True
                    break
        prior_option = {"label": "prior", "value": "prior"}
        if has_priors and prior_option not in plotting_options:
            plotting_options.append(prior_option)
        if not has_priors and prior_option in plotting_options:
            plotting_options.remove(prior_option)
        return plotting_options



import dash
import webviz_ert.models
import webviz_ert.controllers
from dash.development.base_component import Component
from typing import List, Dict
from webviz_ert.views import ensemble_selector_list, response_obs_view
from webviz_ert.plugins import WebvizErtPluginABC


class ObservationAnalyzer(WebvizErtPluginABC):
    def __init__(self, app: dash.Dash, project_identifier: str, beta: bool = False):
        super().__init__(app, project_identifier)
        self.set_callbacks(app)
        self.beta = beta

    @property
    def layout(self) -> Component:
        return dash.html.Div(
            [
                dash.html.Div(
                    children=[
                        dash.html.P(
                            [
                                "This page is considered a ",
                                dash.html.B("beta"),
                                " version and could be changed or removed. You are encouraged to use it and give feedback to us regarding functionality and / or bugs.",
                            ],
                            className="ert-beta-warning",
                            id=self.uuid("beta-warning"),
                        )
                    ],
                    hidden=not self.beta,
                ),
                dash.html.Div(
                    id=self.uuid("ensemble-content"),
                    children=ensemble_selector_list(parent=self),
                ),
                dash.html.Div(
                    id=self.uuid("plotting-content"),
                    children=response_obs_view(parent=self),
                ),
            ]
        )

    @property
    def tour_steps(self) -> List[Dict[str, str]]:
        steps = [
            {
                "id": self.uuid("ensemble-multi-selector"),
                "content": "List of experiment ensembles.",
            },
            {
                "id": self.uuid("selected-ensemble-dropdown"),
                "content": "List of currently selected ensembles.",
            },
            {
                "id": self.uuid(f"ensemble-refresh-button"),
                "content": (
                    "Forces a refresh of all ensemble data including parameter and response data."
                ),
            },
            {
                "id": self.uuid("response-selector"),
                "content": (
                    "Response selection list will be populated"
                    " when an ensemble is selected in the list of ensembles"
                    " and will contain only responses that have observations."
                ),
            },
            {
                "id": self.uuid("observations-graph-container"),
                "content": (
                    "Representation of the misfit between the estimated response value from the forward model"
                    " and the existing observation values."
                    " One can choose to show misfits grouped by time (Univeriate) as candles"
                    " showing misfits statistics per a single time point"
                    " or to render misfits as a histogram aggregating misfits over the entire temporal axes (Summary)."
                    " When multiple ensembles are selected each graph will be overlayed on top of each other"
                    " transparently, where each ensemble gets its own colour"
                ),
            },
        ]
        return steps

    def set_callbacks(self, app: dash.Dash) -> None:
        webviz_ert.controllers.ensemble_list_selector_controller(self, app)
        webviz_ert.controllers.observation_response_controller(self, app)



from typing import List, Dict, Any
from webviz_ert.plugins import WebvizErtPluginABC
import dash
from dash.dependencies import Input, Output, State
from webviz_ert.data_loader import get_ensembles, refresh_data
from webviz_ert.models import load_ensemble


def _get_non_selected_options(store: Dict[str, List]) -> List[Dict[str, str]]:
    _options = []
    for option in store["options"]:
        if option not in store["selected"]:
            _options.append(option)
    return _options


def _setup_ensemble_selection_store(parent: WebvizErtPluginABC) -> Dict[str, List]:
    stored_selection = parent.load_state(key="ensembles", default=[])
    ensemble_selection_store: Dict[str, List] = {"options": [], "selected": []}

    if not parent.get_ensembles():
        ensemble_dict = get_ensembles(project_id=parent.project_identifier)
        for ensemble_schema in ensemble_dict:
            ensemble_id = ensemble_schema["id"]
            load_ensemble(parent, ensemble_id)

    for ens_id, ensemble in parent.get_ensembles().items():
        element = {"label": ensemble.name, "value": ensemble.id}
        ensemble_selection_store["options"].append(element)
        if ensemble.name in stored_selection:
            ensemble_selection_store["selected"].append(element)

    return ensemble_selection_store


def ensemble_list_selector_controller(
    parent: WebvizErtPluginABC, app: dash.Dash
) -> None:
    @app.callback(
        [
            Output(parent.uuid("selected-ensemble-dropdown"), "options"),
            Output(parent.uuid("selected-ensemble-dropdown"), "value"),
            Output(parent.uuid("ensemble-multi-selector"), "options"),
            Output(parent.uuid("ensemble-multi-selector"), "value"),
            Output(parent.uuid("ensemble-selection-store"), "data"),
        ],
        [
            Input(parent.uuid("ensemble-multi-selector"), "value"),
            Input(parent.uuid("selected-ensemble-dropdown"), "value"),
            Input(parent.uuid(f"ensemble-refresh-button"), "n_clicks"),
        ],
        [
            State(parent.uuid("selected-ensemble-dropdown"), "options"),
            State(parent.uuid("selected-ensemble-dropdown"), "value"),
            State(parent.uuid("ensemble-multi-selector"), "options"),
            State(parent.uuid("ensemble-selection-store"), "data"),
        ],
    )
    def set_callback(
        _input_ensemble_selector: List[str],
        _: List[str],
        _btn_click: int,
        selected_ens_options: List[Dict],
        selected_ens_value: List[str],
        ens_selector_options: List[Dict[str, str]],
        ensemble_selection_store: Dict[str, List],
    ) -> List[Any]:
        ctx = dash.callback_context
        triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]
        if not triggered_id and not ensemble_selection_store:
            ensemble_selection_store = _setup_ensemble_selection_store(parent)

        if triggered_id == parent.uuid("ensemble-multi-selector"):
            selected_list_elem = _input_ensemble_selector[0]
            element = next(
                op for op in ens_selector_options if op["value"] == selected_list_elem
            )
            ensemble_selection_store["selected"].append(element)
            parent.save_state(
                "ensembles",
                [el["label"] for el in ensemble_selection_store["selected"]],
            )

        if triggered_id == parent.uuid("selected-ensemble-dropdown"):
            for element in [
                op
                for op in selected_ens_options
                if op["value"] not in selected_ens_value
            ]:
                ensemble_selection_store["selected"].remove(element)
            parent.save_state(
                "ensembles",
                [el["label"] for el in ensemble_selection_store["selected"]],
            )

        if triggered_id == parent.uuid(f"ensemble-refresh-button"):
            parent.clear_ensembles()
            refresh_data(project_id=parent.project_identifier)
            parent.save_state("ensembles", [])
            ensemble_selection_store = _setup_ensemble_selection_store(parent)

        ens_selector_options = _get_non_selected_options(ensemble_selection_store)
        selected_ens_options = ensemble_selection_store["selected"]
        selected_ens_value = [option["value"] for option in selected_ens_options]

        return [
            selected_ens_options,
            selected_ens_value,
            ens_selector_options,
            [],
            ensemble_selection_store,
        ]

    container_ensemble_selector_multi_id = parent.uuid(
        "container-ensemble-selector-multi"
    )
    parameter_selector_button_id = parent.uuid(f"parameter-selector-button")

    @app.callback(
        [
            Output(container_ensemble_selector_multi_id, "className"),
            Output(parameter_selector_button_id, "children"),
        ],
        [
            Input(parameter_selector_button_id, "n_clicks"),
            Input(parameter_selector_button_id, "children"),
        ],
        [
            State(container_ensemble_selector_multi_id, "className"),
        ],
    )
    def toggle_selector_visibility(
        _: int, button_text: str, class_name: str
    ) -> List[str]:
        ctx = dash.callback_context
        triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]
        if triggered_id == parameter_selector_button_id:
            if class_name == "ert-ensemble-selector-container-hide":
                class_name = "ert-ensemble-selector-container-show"
                button_text = "Hide Selectors"
            else:
                class_name = "ert-ensemble-selector-container-hide"
                button_text = "Show Selectors"
        return [class_name, button_text]



import pytest
from selenium.webdriver.common.keys import Keys

from tests.conftest import select_by_name, select_ensemble, setup_plugin
from webviz_ert.plugins import ParameterComparison


@pytest.mark.browser_test
def test_parameter_selector(
    mock_data,
    dash_duo,
):
    plugin = setup_plugin(dash_duo, __name__, ParameterComparison)
    select_ensemble(dash_duo, plugin)

    dash_duo.wait_for_element("#" + plugin.uuid("parameter-selector-multi-param"))

    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("parameter-selector-multi-param"),
        "BPR_138_PERSISTENCE",
    )

    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("parameter-selector-multi-param"),
        "OP1_DIVERGENCE_SCALE",
    )

    paremeter_deactivator = dash_duo.find_element(
        "#" + plugin.uuid("parameter-deactivator-param")
    )

    paremeter_deactivator.click()

    parameter_selector_input = dash_duo.find_element(
        "#" + plugin.uuid("parameter-selector-filter-param")
    )

    parameter_selector_input.send_keys("OP1")

    dash_duo.wait_for_text_to_equal(
        "#" + plugin.uuid("parameter-selector-filter-param"), "OP1"
    )
    dash_duo.wait_for_text_to_equal(
        "#" + plugin.uuid("parameter-selector-multi-param"),
        "OP1_DIVERGENCE_SCALE",
    )

    parameter_selector_container = dash_duo.find_element(
        "#" + plugin.uuid("container-parameter-selector-multi-param")
    )

    assert parameter_selector_container.is_displayed()
    button_hide = dash_duo.find_element("#" + plugin.uuid("parameter-selector-button"))
    button_hide.click()
    parameter_selector_container = dash_duo.wait_for_element_by_css_selector(
        ".ert-parameter-selector-container-hide"
    )
    # assert dash_duo.get_logs() == []


@pytest.mark.browser_test
def test_search_input_return_functionality(
    mock_data,
    dash_duo,
):
    plugin = setup_plugin(dash_duo, __name__, ParameterComparison)
    select_ensemble(dash_duo, plugin)

    parameter_selector_container = dash_duo.find_element(
        "#" + plugin.uuid("container-parameter-selector-multi-param")
    )

    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("parameter-selector-multi-param"),
        "BPR_138_PERSISTENCE",
    )
    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("parameter-selector-multi-param"),
        "OP1_DIVERGENCE_SCALE",
    )
    first_elem, _ = parameter_selector_container.text.split("\n")

    param_selector_id = plugin.uuid("parameter-selector-multi-param")
    select_by_name(dash_duo, f"#{param_selector_id}", "BPR_138_PERSISTENCE")

    parameter_deactivator = dash_duo.find_element(
        "#" + plugin.uuid("parameter-deactivator-param")
    )

    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("parameter-deactivator-param"),
        "×BPR_138_PERSISTENCE",
    )
    parameter_deactivator.click()
    dash_duo.clear_input(parameter_deactivator)

    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("parameter-deactivator-param"), ""
    )

    parameter_selector_input = dash_duo.find_element(
        "#" + plugin.uuid("parameter-selector-filter-param")
    )
    parameter_selector_input.send_keys(Keys.ENTER)
    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("parameter-deactivator-param"), ""
    )

    parameter_selector_input.send_keys("OP1")

    dash_duo.wait_for_text_to_equal(
        "#" + plugin.uuid("parameter-selector-filter-param"), "OP1"
    )
    dash_duo.wait_for_text_to_equal(
        "#" + plugin.uuid("parameter-selector-multi-param"),
        "OP1_DIVERGENCE_SCALE",
    )
    parameter_selector_input.send_keys(Keys.ENTER)

    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("parameter-deactivator-param"), ""
    )

    select_by_name(dash_duo, f"#{param_selector_id}", "OP1_DIVERGENCE_SCALE")
    dash_duo.wait_for_contains_text(
        "#" + plugin.uuid("parameter-deactivator-param"),
        "×OP1_DIVERGENCE_SCALE",
    )

    # assert dash_duo.get_logs() == []


@pytest.mark.browser_test
def test_parameter_selector_sorting(
    mock_data,
    dash_duo,
):
    plugin = setup_plugin(dash_duo, __name__, ParameterComparison)

    wanted_ensemble_name = "nr_42"
    select_ensemble(dash_duo, plugin, wanted_ensemble_name)

    expected_parameters = [
        "BPR_138_PERSISTENCE",
        "OP1_DIVERGENCE_SCALE",
    ]

    parameter_selector_container_id = "#" + plugin.uuid(
        "container-parameter-selector-multi-param"
    )
    dash_duo.wait_for_text_to_equal(
        parameter_selector_container_id, "\n".join(expected_parameters)
    )



from dash.development.base_component import Component
from webviz_config import WebvizPluginABC

from dash import html
from dash import dcc


def correlation_view(id_view: str) -> Component:
    return html.Div(
        id=f"container_{id_view}",
        className="ert-view-container",
        children=[
            dcc.Graph(
                id=id_view,
                className="ert-view-cell",
                config={"responsive": True},
            )
        ],
    )



from webviz_ert.models.ensemble_model import EnsembleModel
from webviz_ert.controllers.controller_functions import (
    response_options,
    _valid_response_option,
)


def test_response_options(mock_data):
    ensemble3 = EnsembleModel(ensemble_id=3, project_id=None)
    ensemble4 = EnsembleModel(ensemble_id=4, project_id=None)

    options = response_options(response_filters=[], ensembles=[ensemble3, ensemble4])
    options_with_obs = response_options(
        response_filters=["obs"], ensembles=[ensemble3, ensemble4]
    )

    assert len(options) == 4
    assert len(options_with_obs) == 2


def test_valid_response_option():
    response_filters = ["obs", "historical"]

    class Response:
        name = "DummyResponse"
        has_observations = True

    response = Response()
    assert _valid_response_option(response_filters, response) is True

    response.name = "DummyResponseH"
    assert _valid_response_option(response_filters, response) is False

    response.name = "DummyResponseH:OP_1"
    assert _valid_response_option(response_filters, response) is False

    response.has_observations = False
    assert _valid_response_option(response_filters, response) is False

    response.name = "DummyResponse"
    assert _valid_response_option(response_filters, response) is False

    response_filters = ["obs"]
    response.name = "DummyResponseH"
    response.has_observations = True
    assert _valid_response_option(response_filters, response) is True

    response_filters = ["obs"]
    response.name = "DummyResponseH"
    response.has_observations = False
    assert _valid_response_option(response_filters, response) is False

    response_filters = ["historical"]
    response.name = "DummyResponse"
    response.has_observations = False
    assert _valid_response_option(response_filters, response) is True

    response_filters = ["historical"]
    response.name = "DummyResponseH"
    response.has_observations = False
    assert _valid_response_option(response_filters, response) is False



from typing import Dict, Optional, List, Union
import pandas as pd
import datetime

from webviz_ert.models import indexes_to_axis, AxisType


class Observation:
    def __init__(self, observation_schema: Dict):
        self.name = str(observation_schema["name"])
        self._x_axis = indexes_to_axis(observation_schema["x_axis"])
        self._std = observation_schema["errors"]
        self._values = observation_schema["values"]
        self._attributes = ""
        self._active = [True for _ in self._x_axis] if self._x_axis else []

        if "attributes" in observation_schema:
            for k, v in observation_schema["attributes"].items():
                self._attributes += f"{k}: {v}<br>"

    def data_df(self) -> pd.DataFrame:
        return pd.DataFrame(
            data={
                "values": self._values,
                "std": self._std,
                "x_axis": indexes_to_axis(self._x_axis),
                "attributes": self._attributes,
                "active": self._active,
            }
        )

    @property
    def axis(self) -> Optional[List[Union[int, str, datetime.datetime]]]:
        return self._x_axis

    @property
    def axis_type(self) -> Optional[AxisType]:
        if self.axis is None:
            return None
        if str(self.axis[0]).isnumeric():
            return AxisType.INDEX
        return AxisType.TIMESTAMP



import sys

if sys.version_info < (3, 11):
    from enum import Enum

    class StrEnum(str, Enum):
        pass

else:
    from enum import StrEnum


class DataType(StrEnum):
    RESPONSE = "resp"
    PARAMETER = "param"
    ENSEMBLE = "ens"


class AxisType(StrEnum):
    INDEX = "index"
    TIMESTAMP = "timestamp"



import pytest


@pytest.fixture(scope="session", autouse=True)
def start_storage_server():
    try:
        from ert.services import StorageService as Storage
    except ImportError:
        from ert.services import Storage

    with Storage.start_server() as service:
        service.wait_until_ready(timeout=30)
        yield service



import pytest

from tests.conftest import (
    select_ensemble,
    select_parameter,
    select_response,
    setup_plugin,
    wait_a_bit,
)
from webviz_ert.plugins._response_comparison import ResponseComparison


@pytest.mark.browser_test
def test_plot_view(
    mock_data,
    dash_duo,
):
    # This test selects an ensemble from the ensemble-multi-selector
    # then selects a response and parameter and checks that the
    # DOM element for both are created.
    plugin = setup_plugin(dash_duo, __name__, ResponseComparison)

    select_ensemble(dash_duo, plugin)

    select_response(dash_duo, plugin, "SNAKE_OIL_GPR_DIFF@199")

    select_parameter(dash_duo, plugin, "BPR_138_PERSISTENCE")

    # assert dash_duo.get_logs() == [], "browser console should contain no error"


@pytest.mark.browser_test
def test_clearing_parameters_view(
    mock_data,
    dash_duo,
):
    # this is a regression test, ensuring that all appropriate plots are
    # cleared when a selection of multiple parameters is cleared.
    # we select an ensemble, select a response, select two parameters, check
    # that three graphs are present, clear all parameters, and check that just
    # the response graph remains
    plugin = setup_plugin(dash_duo, __name__, ResponseComparison)

    select_ensemble(dash_duo, plugin)

    response_name = "SNAKE_OIL_GPR_DIFF@199"

    # click the response, and check that the response graph appears
    select_response(dash_duo, plugin, response_name)

    # click some parameters
    select_parameter(dash_duo, plugin, "BPR_138_PERSISTENCE")
    select_parameter(dash_duo, plugin, "OP1_DIVERGENCE_SCALE")

    # clear parameter selection
    clear_all = dash_duo.find_element(
        "#" + plugin.uuid("parameter-deactivator-param") + " span.Select-clear-zone"
    )
    clear_all.click()

    # verify only expected response plot is left in place
    dash_duo.find_element(f'.dash-graph[id*="{response_name}"]')

    # assert dash_duo.get_logs() == [], "browser console sho uld contain no error"


@pytest.mark.browser_test
def test_clearing_ensembles_view(
    mock_data,
    dash_duo,
):
    # this is a regression test, ensuring that clearing all ensembles actually
    # removes all ensembles, all responses, all parameters, and all plots

    plugin = setup_plugin(dash_duo, __name__, ResponseComparison, (1024, 2048))

    # click and choose two ensembles
    select_ensemble(dash_duo, plugin)
    select_ensemble(dash_duo, plugin)

    # click the response, and check that the response graph appears
    select_response(dash_duo, plugin, "SNAKE_OIL_GPR_DIFF@199")

    # click some parameters
    select_parameter(dash_duo, plugin, "BPR_138_PERSISTENCE")
    select_parameter(dash_duo, plugin, "OP1_DIVERGENCE_SCALE")

    # clear ensemble selection
    clear_all = dash_duo.find_element(
        "#" + plugin.uuid("selected-ensemble-dropdown") + " span.Select-clear-zone"
    )
    clear_all.click()

    # wait a bit for the page to update
    wait_a_bit(dash_duo)

    # verify all plots are gone
    plots = dash_duo.find_elements(".dash-graph")
    assert len(plots) == 0

    # verify no responses are selected
    chosen_responses = dash_duo.find_elements(
        "#" + plugin.uuid("parameter-deactivator-resp") + " .Select-value"
    )
    assert len(chosen_responses) == 0

    # verify no parameters are selected
    chosen_parameters = dash_duo.find_elements(
        "#" + plugin.uuid("parameter-deactivator-param") + " .Select-value"
    )
    assert len(chosen_parameters) == 0

    # assert dash_duo.get_logs() == [], "browser console should contain no error"


@pytest.mark.browser_test
def test_axis_labels(mock_data, dash_duo):
    """test_axis_labels loads two different plots in the plot view and checks
    that axes are labelled correctly"""

    plugin = setup_plugin(
        dash_duo, __name__, ResponseComparison, window_size=(1024, 2048)
    )

    # find the right ensemble which has mock data prepared for this test
    wanted_ensemble_name = "default3"
    select_ensemble(dash_duo, plugin, wanted_ensemble_name)

    # click two responses, with specified names
    wanted_responses = ["FGPT", "FOPR"]

    for response in wanted_responses:
        select_response(dash_duo, plugin, response)

    # check that both have Value as y axis label
    for response in wanted_responses:
        y_axis_title_selector = f"#{plugin.uuid(response)} text.ytitle"
        dash_duo.wait_for_text_to_equal(y_axis_title_selector, "Value")

    # check that one has date, the other has index as x axis label
    date_plot_id = plugin.uuid("FOPR")
    dash_duo.wait_for_text_to_equal(f"#{date_plot_id} text.xtitle", "Date")

    index_plot_id = plugin.uuid("FGPT")
    dash_duo.wait_for_text_to_equal(f"#{index_plot_id} text.xtitle", "Index")

    # assert dash_duo.get_logs() == [], "browser console should contain no error"



from typing import List, Any, Dict, Mapping, Optional, TYPE_CHECKING, Union
import numpy as np
import math
import pandas as pd
import logging
import plotly.graph_objects as go
import plotly.figure_factory as ff
from dateutil.parser import isoparse
from scipy.stats import norm, lognorm, truncnorm, uniform, loguniform, triang
import webviz_ert.assets as assets
from webviz_ert.models.data_model import AxisType

if TYPE_CHECKING:
    from webviz_ert.models.parameter_model import PriorModel
"""
Unsupported priors
CONST
DUNIF
ERRF
DERRF
"""

logger = logging.getLogger()

WARNING_MSG = """Plotting {prior_type} is not yet fully supported.
Please contact us on slack channel #ert-users if this is a desired feature."""


def _TRIANGULAR(
    xaxis: List[float], xmin: float, xmode: float, xmax: float
) -> np.ndarray:
    loc = xmin
    scale = xmax - xmin
    c = (xmode - xmin) / scale
    return triang.pdf(xaxis, c, loc, scale)


def _TRUNC_NORMAL(
    xaxis: List[float], mean: float, std: float, _min: float, _max: float
) -> np.ndarray:
    logger.warning(WARNING_MSG.format(prior_type="TRUNCATED_NORMAL"))
    a = (_min - mean) / std
    b = (_max - mean) / std
    return truncnorm.pdf(xaxis, a, b, mean, std)


def _CONST(xaxis: List[float], value: float) -> List[float]:
    logger.warning(WARNING_MSG.format(prior_type="CONST"))
    return [value for x in xaxis]


def _UNIFORM(xaxis: List[float], _min: float, _max: float) -> np.ndarray:
    loc = _min
    scale = _max - _min
    return uniform.pdf(xaxis, loc, scale)


def _DUNIFORM(xaxis: List[float], _min: float, _max: float) -> np.ndarray:
    logger.warning(WARNING_MSG.format(prior_type="DUNIF"))
    loc = _min
    scale = _max - _min
    return uniform.pdf(xaxis, loc, scale)


def _RAW(xaxis: List[float]) -> np.ndarray:
    return norm.pdf(xaxis, 0, 1)


def _ERRF(*args: Any) -> List:
    logger.warning(WARNING_MSG.format(prior_type="ERRF"))
    return []


def _DERRF(*args: Any) -> List:
    logger.warning(WARNING_MSG.format(prior_type="DERRF"))
    return []


PRIOR_FUNCTIONS = {
    "normal": norm.pdf,
    "lognormal": lognorm.pdf,
    "ert_truncnormal": _TRUNC_NORMAL,
    "uniform": _UNIFORM,
    "DUNIF": _DUNIFORM,
    "loguniform": loguniform.pdf,
    "trig": _TRIANGULAR,
    "const": _CONST,
    "stdnormal": _RAW,
    "ert_ert": _ERRF,
    "ert_derf": _DERRF,
}


def _create_prior_plot(
    name: str, prior: "PriorModel", _min: float, _max: float, color: str
) -> go.Scattergl:
    n = 100
    diff = (_max - _min) / n
    xaxis = [_min + i * diff for i in range(n)]
    yaxis = PRIOR_FUNCTIONS[prior.function](xaxis, *prior.function_parameter_values)
    style = assets.ERTSTYLE["parameter-plot"]["prior"].copy()
    style["line"]["color"] = color

    return go.Scattergl(name=f"{name}-prior", x=xaxis, y=yaxis, line=style["line"])


class BoxPlotModel:
    def __init__(self, **kwargs: Any):
        self.selected = True
        self._y_axis = kwargs["y_axis"]
        self._name = kwargs["name"]
        self._color = kwargs["color"]
        self._customdata = kwargs["customdata"] if "customdata" in kwargs else None
        self._hovertemplate = (
            kwargs["hovertemplate"] if "hovertemplate" in kwargs else None
        )
        self._ensemble_name = (
            kwargs["ensemble_name"] if "ensemble_name" in kwargs else None
        )

    @property
    def repr(self) -> go.Box:
        repr_dict = dict(
            y=self._y_axis,
            x0=f"{self._name}",
            legendgroup=self.name,
            name=self.display_name,
            boxpoints="all",
            jitter=0.3,
            pointpos=-1.8,
            marker_color=self._color,
            customdata=self._customdata,
            hovertemplate=self._hovertemplate,
        )

        return go.Box(repr_dict)

    @property
    def name(self) -> str:
        return self._name

    @property
    def display_name(self) -> str:
        if isinstance(self._name, int):
            return f"Value {self._name}"
        else:
            return (
                f"{self._name} - {self._ensemble_name}"
                if self._ensemble_name
                else self._name
            )


class BarChartPlotModel:
    def __init__(
        self, data_df_dict: Mapping[str, pd.DataFrame], colors: Mapping[str, str]
    ):
        self._data_df_dict = data_df_dict
        self._colors = colors
        self.selection: List[int] = []

    @property
    def plot_ids(self) -> Mapping[int, str]:
        return {plot_idx: ens_name for plot_idx, ens_name in enumerate(self.data)}

    @property
    def data(self) -> Mapping[str, pd.DataFrame]:
        return self._data_df_dict

    @property
    def repr(self) -> go.Figure:
        fig = go.Figure()
        for ens_name in self.data:
            fig.add_trace(
                go.Bar(
                    x=self.data[ens_name].values,
                    y=self.data[ens_name].index,
                    orientation="h",
                    name=ens_name,
                    marker=dict(
                        color=self._colors[ens_name],
                    ),
                ),
            )
        _layout = assets.ERTSTYLE["figure"]["layout"].copy()
        _layout.update(
            {
                "showlegend": False,
                "yaxis": {
                    "showticklabels": True,
                },
                "font": {"size": 10},
            }
        )
        fig.update_layout(_layout)
        return fig


class PlotModel:
    def __init__(self, **kwargs: Any):
        self._x_axis = kwargs["x_axis"]
        self._y_axis = kwargs["y_axis"]
        self._text = kwargs.get("text")
        self._name = kwargs["name"]
        self._mode = kwargs["mode"]
        self._line = kwargs["line"]
        self._marker = kwargs["marker"]
        self._error_y = kwargs.get("error_y")
        self._hoverlabel = kwargs.get("hoverlabel")
        self._meta = kwargs.get("meta")
        self._xaxis = kwargs.get("xaxis")
        self.selected = True
        self.legendgroup = kwargs.get("legendgroup")
        self.showlegend = kwargs.get("showlegend", True)

    @property
    def repr(self) -> Union[go.Scattergl, go.Scatter]:
        repr_dict = dict(
            x=self._x_axis,
            y=self._y_axis,
            text=self._text,
            name=self.display_name,
            mode=self._mode,
            error_y=self._error_y,
            xaxis=self._xaxis,
            connectgaps=True,
            hoverlabel=self._hoverlabel,
            meta=self._meta,
            showlegend=self.showlegend,
        )
        if self.legendgroup:
            repr_dict["legendgroup"] = self.legendgroup
        if self._line:
            repr_dict["line"] = self._line
        if self._marker:
            repr_dict["marker"] = self._marker
            repr_dict["unselected"] = {"marker": {"size": 9}}
        if self.selected:
            repr_dict["visible"] = True
        else:
            repr_dict["visible"] = "legendonly"

        return go.Scattergl(repr_dict)

    @property
    def name(self) -> str:
        return self._name

    @property
    def display_name(self) -> str:
        if isinstance(self._name, int):
            return f"Realization {self._name}"
        else:
            return self._name

    @property
    def axis_type(self) -> Any:
        if str(self._x_axis[0]).isnumeric():
            return AxisType.INDEX
        return AxisType.TIMESTAMP

    @property
    def x_axis(self) -> Any:
        return self._x_axis

    def indexes_in_range(self, x_range: List) -> List:
        result = []
        if self.axis_type == AxisType.TIMESTAMP:
            x_range = [pd.Timestamp(x) for x in x_range]
        x_start = min(x_range)
        x_end = max(x_range)
        for val in self.x_axis:
            if x_start <= val <= x_end:
                result.append(val)
        return result


class ResponsePlotModel:
    def __init__(
        self,
        realization_plots: List[Any],
        observations: List[Any],
        layout: Dict,
    ):
        self._realization_plots = realization_plots
        self._observations = observations
        self._layout = layout
        self.selection: List[int] = []

    @property
    def repr(self) -> go.Figure:
        if self.selection:
            for real in self._realization_plots:
                real.selected = real.name in self.selection
        else:
            for real in self._realization_plots:
                real.selected = True

        fig = go.Figure(layout=self._layout)
        for rel in self._realization_plots:
            fig.add_trace(rel.repr)
        for obs in self._observations:
            fig.add_trace(obs.repr)
        return fig

    @property
    def plot_ids(self) -> Mapping[int, str]:
        return {idx: rel.name for idx, rel in enumerate(self._realization_plots)}


class MultiHistogramPlotModel:
    def __init__(
        self,
        data_df_dict: Mapping[str, pd.DataFrame],
        names: Mapping[str, str],
        colors: Mapping[str, str],
        hist: bool = True,
        kde: bool = True,
        priors: Dict = {},
        bin_count: Optional[int] = None,
    ):
        self._hist_enabled = hist
        self._kde_enabled = kde
        self._data_df_dict = data_df_dict
        self.selection: List[int] = []
        self._colors = colors
        self._priors = priors
        self._bin_count = bin_count
        self._names = names

    @property
    def data_df(self) -> Mapping[str, pd.DataFrame]:
        if self.selection:
            return {
                response_name: self._data_df_dict[response_name][self.selection]
                for response_name in self._data_df_dict
            }
        return self._data_df_dict

    @property
    def plot_ids(self) -> Mapping[int, str]:
        df = list(self._data_df_dict.values())[0]
        return {plot_idx: real.name for plot_idx, real in enumerate(df)}

    @property
    def bin_count(self) -> int:
        if self._bin_count is None:
            # get any dict response data_frame, so the first is a good choice
            data = list(self._data_df_dict.values())[0].values.flatten()
            self._bin_count = int(math.ceil(math.sqrt(len(data)))) - 1
        return self._bin_count

    @property
    def repr(self) -> go.Figure:
        colors = []
        data = []
        names = []
        realization_nums = []
        for name in self._data_df_dict:
            if self._data_df_dict[name].empty:
                continue
            data.append(list(self._data_df_dict[name].values.flatten()))
            colors.append(self._colors[name])
            names.append(self._names[name])
            realization_nums.append(
                [f"Realization {num}" for num in self._data_df_dict[name].columns]
            )
        if data == []:
            return go.Figure()
        _max = float(np.hstack(data).max())
        _min = float(np.hstack(data).min())
        try:
            bin_size = float((_max - _min) / self.bin_count)
        except ZeroDivisionError:
            bin_size = 1.0

        fig = ff.create_distplot(
            data,
            names,
            show_hist=self._hist_enabled,
            show_curve=self._kde_enabled,
            bin_size=bin_size,
            colors=colors,
            show_rug=True,
            rug_text=realization_nums,
            curve_type="normal",
            histnorm="probability density",
        )
        for name, (prior, color) in self._priors.items():
            fig.add_trace(_create_prior_plot(name, prior, _min, _max, color))
        fig.update_layout(clickmode="event+select")
        fig.update_layout(assets.ERTSTYLE["figure"]["layout"])
        return fig


class ParallelCoordinatesPlotModel:
    def __init__(
        self, data_df_dict: Mapping[str, pd.DataFrame], colors: Mapping[str, str]
    ):
        self._data_df_dict = data_df_dict
        self._colors = colors
        self.selection: List[int] = []

    @property
    def plot_ids(self) -> Mapping[int, str]:
        return {plot_idx: parameter for plot_idx, parameter in enumerate(self.data)}

    @property
    def data(self) -> Mapping[str, pd.DataFrame]:
        return self._data_df_dict

    @property
    def repr(self) -> go.Figure:
        fig = go.Figure()
        # rearange colors
        colors = []
        for idx, ens in enumerate(self._colors):
            colors.append([idx / len(self._colors), self._colors[ens]])
            colors.append([(idx + 1) / len(self._colors), self._colors[ens]])
        # create a single dataframe, where all ensemble parameters are stacked
        data_df = pd.concat(list(self.data.values()))
        # drop a column with ids
        ensemble_ids = data_df["ensemble_id"]
        data_df = data_df.drop("ensemble_id", axis=1)
        # create parallel coordinates dimension list
        dimensions = []
        for parameter in data_df:
            dimensions.append(dict(label=parameter, values=data_df[parameter]))
        fig.add_trace(
            go.Parcoords(
                line=dict(
                    color=ensemble_ids,
                    colorscale=colors,
                    showscale=True,
                    colorbar=dict(
                        tickvals=list(range(0, len(self.data))),
                        ticktext=list(self.data.keys()),
                        len=0.2 * len(self.data),
                        title="Ensemble",
                    ),
                ),
                dimensions=dimensions,
                labelangle=45,
                labelside="bottom",
            )
        )
        fig.update_layout(clickmode="event+select")
        fig.update_layout(uirevision=True)

        return fig


class HistogramPlotModel:
    def __init__(self, data_df: pd.DataFrame, hist: bool = True, kde: bool = True):
        self._hist_enabled = hist
        self._kde_enabled = kde
        self._data_df = data_df
        self.selection: List[int] = []

    @property
    def data_df(self) -> pd.DataFrame:
        if self.selection:
            return self._data_df[self.selection]
        return self._data_df

    @property
    def plot_ids(self) -> Mapping[int, str]:
        return {plot_idx: real.name for plot_idx, real in enumerate(self._data_df)}

    @property
    def repr(self) -> go.Figure:
        bin_count = int(math.ceil(math.sqrt(len(self.data_df.values.flatten()))))
        bin_size = float(
            (
                (self.data_df.max(axis=1).values - self.data_df.min(axis=1).values)
                / bin_count
            )[0]
        )
        fig = ff.create_distplot(
            [list(self.data_df.values.flatten())],
            [self.data_df.index.name],
            show_hist=self._hist_enabled,
            show_curve=self._kde_enabled,
            bin_size=bin_size,
        )
        fig.update_layout(clickmode="event+select")

        fig.update_layout(assets.ERTSTYLE["figure"]["layout"])
        return fig



import dash
import pytest

from tests.conftest import (
    select_ensemble,
    select_parameter,
    select_response,
    setup_plugin,
)
from webviz_ert.plugins import ResponseComparison, WebvizErtPluginABC


@pytest.mark.browser_test
def test_state_saved(mock_data, dash_duo, tmpdir):
    root_path = tmpdir.strpath
    plugin = setup_plugin(
        dash_duo, __name__, ResponseComparison, project_identifier=root_path
    )

    ens_name = "default"
    resp_name = "SNAKE_OIL_GPR_DIFF@199"
    param_name = "BPR_138_PERSISTENCE"

    select_ensemble(dash_duo, plugin, ens_name)
    select_response(dash_duo, plugin, resp_name)
    select_parameter(dash_duo, plugin, param_name)

    # Check state has been saved
    plugin_state = plugin.load_state()
    assert plugin_state[f"{plugin._class_name}"]["ensembles"] == [ens_name]
    assert plugin_state[f"{plugin._class_name}"]["param"] == [param_name]
    assert plugin_state[f"{plugin._class_name}"]["resp"] == [resp_name]

    state_before_quit = plugin_state

    # Simulate application exit by killing the driver
    dash_duo.driver.quit()
    # Manually reset the global state
    WebvizErtPluginABC._state = {}
    # Check plugin state is also reset
    assert plugin.load_state() == {}

    # Check initializing new plugin will load state from disk if state is not set
    app = dash.Dash(__name__)
    new_plugin = ResponseComparison(app, project_identifier=root_path)
    state_after_loading = new_plugin.load_state()

    assert state_after_loading == state_before_quit



from webviz_config import WebvizPluginABC
from dash.development.base_component import Component

from dash import html
from dash import dcc
import webviz_ert.assets as assets


def parallel_coordinates_view(parent: WebvizPluginABC) -> Component:
    return html.Div(
        className="ert-view-container",
        children=[
            dcc.Graph(
                id={
                    "id": parent.uuid("parallel-coor"),
                    "type": parent.uuid("graph"),
                },
                className="ert-view-cell allow-overflow",
                config={"responsive": True},
            )
        ],
    )



import dash
import dash_bootstrap_components as dbc
import webviz_ert.controllers
import webviz_ert.models
from dash.development.base_component import Component
from typing import List, Dict

from webviz_ert.views import (
    ensemble_selector_list,
    parallel_coordinates_view,
    parameter_selector_view,
)
from webviz_ert.plugins import WebvizErtPluginABC
from webviz_ert.models.data_model import DataType


class ParameterComparison(WebvizErtPluginABC):
    def __init__(self, app: dash.Dash, project_identifier: str, beta: bool = False):
        super().__init__(app, project_identifier)
        self.set_callbacks(app)
        self.beta = beta

    @property
    def layout(self) -> Component:
        return dash.html.Div(
            [
                dash.html.Div(
                    children=[
                        dash.html.P(
                            [
                                "This page is considered a ",
                                dash.html.B("beta"),
                                " version and could be changed or removed. You are encouraged to use it and give feedback to us regarding functionality and / or bugs.",
                            ],
                            className="ert-beta-warning",
                            id=self.uuid("beta-warning"),
                        )
                    ],
                    hidden=not self.beta,
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            id=self.uuid("ensemble-content"),
                            children=ensemble_selector_list(parent=self),
                            width=6,
                        ),
                        dbc.Col(
                            id=self.uuid("parameter-content"),
                            children=[
                                parameter_selector_view(
                                    parent=self, data_type=DataType.PARAMETER
                                ),
                            ],
                            width=6,
                        ),
                    ]
                ),
                dbc.Row(
                    dbc.Col(
                        id=self.uuid("parallel-coor-content"),
                        children=[
                            parallel_coordinates_view(parent=self),
                        ],
                        width=12,
                    )
                ),
            ]
        )

    @property
    def tour_steps(self) -> List[Dict[str, str]]:
        steps = [
            {
                "id": self.uuid("ensemble-multi-selector"),
                "content": "List of experiment ensembles.",
            },
            {
                "id": self.uuid("selected-ensemble-dropdown"),
                "content": "List of currently selected ensembles.",
            },
            {
                "id": self.uuid(f"ensemble-refresh-button"),
                "content": (
                    "Forces a refresh of all ensemble data including parameter and response data."
                ),
            },
            {
                "id": self.uuid(f"parameter-selector-multi-param"),
                "content": (
                    "List of parameters. This list is populated only"
                    " if at least one ensemble is selected."
                    " Selecting multiple parameters is possible"
                    " using mouse `Click + Drag` inside the response list."
                ),
            },
            {
                "id": self.uuid(f"parameter-deactivator-param"),
                "content": (
                    "List of currently selected parameters."
                    "Every selected parameter is visualized in the parallel coordinates plot,"
                    " where each vertical axis represents a single parameter."
                    " Multiple ensembles are distinguished by different colors."
                ),
            },
            {
                "id": self.uuid(f"parameter-selector-filter-param"),
                "content": (
                    "Search field. The parameter list will show only"
                    " elements that contain the search characters"
                ),
            },
        ]
        return steps

    def set_callbacks(self, app: dash.Dash) -> None:
        webviz_ert.controllers.ensemble_list_selector_controller(self, app)
        webviz_ert.controllers.parameter_selector_controller(
            self, app, data_type=DataType.PARAMETER, union_keys=False
        )
        webviz_ert.controllers.parameter_comparison_controller(self, app)
